# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase


class DeleteArtifactCollectionTagAssignments(GQLBase):
    delete_artifact_collection_tag_assignments: Optional[
        DeleteArtifactCollectionTagAssignmentsDeleteArtifactCollectionTagAssignments
    ] = Field(alias="deleteArtifactCollectionTagAssignments")


class DeleteArtifactCollectionTagAssignmentsDeleteArtifactCollectionTagAssignments(
    GQLBase
):
    success: bool


DeleteArtifactCollectionTagAssignments.model_rebuild()



"""Scopes in which a W&B Automation can be triggered."""

from __future__ import annotations

from typing import Literal, Union

from pydantic import BeforeValidator, Field
from typing_extensions import Annotated, TypeAlias, get_args

from wandb._pydantic import GQLBase
from wandb.automations._generated import (
    ArtifactPortfolioScopeFields,
    ArtifactSequenceScopeFields,
    ProjectScopeFields,
)

from ._validators import LenientStrEnum, to_scope


# NOTE: Re-defined publicly with a more readable name for easier access
class ScopeType(LenientStrEnum):
    """The kind of scope that triggers an automation."""

    PROJECT = "PROJECT"
    ARTIFACT_COLLECTION = "ARTIFACT_COLLECTION"


class _BaseScope(GQLBase):
    scope_type: Annotated[ScopeType, Field(frozen=True)]


class _ArtifactSequenceScope(_BaseScope, ArtifactSequenceScopeFields):
    """An automation scope defined by a specific `ArtifactSequence`."""

    scope_type: Literal[ScopeType.ARTIFACT_COLLECTION] = ScopeType.ARTIFACT_COLLECTION


class _ArtifactPortfolioScope(_BaseScope, ArtifactPortfolioScopeFields):
    """An automation scope defined by a specific `ArtifactPortfolio` (e.g. a registry collection)."""

    scope_type: Literal[ScopeType.ARTIFACT_COLLECTION] = ScopeType.ARTIFACT_COLLECTION


# for type annotations
ArtifactCollectionScope = Annotated[
    Union[_ArtifactSequenceScope, _ArtifactPortfolioScope],
    BeforeValidator(to_scope),
    Field(discriminator="typename__"),
]
"""An automation scope defined by a specific `ArtifactCollection`."""

# for runtime type checks
ArtifactCollectionScopeTypes: tuple[type, ...] = get_args(
    ArtifactCollectionScope.__origin__  # type: ignore[attr-defined]
)


class ProjectScope(_BaseScope, ProjectScopeFields):
    """An automation scope defined by a specific `Project`."""

    scope_type: Literal[ScopeType.PROJECT] = ScopeType.PROJECT


# for type annotations
AutomationScope: TypeAlias = Annotated[
    Union[_ArtifactSequenceScope, _ArtifactPortfolioScope, ProjectScope],
    BeforeValidator(to_scope),
    Field(discriminator="typename__"),
]
# for runtime type checks
AutomationScopeTypes: tuple[type, ...] = get_args(AutomationScope.__origin__)  # type: ignore[attr-defined]


__all__ = [
    "ScopeType",
    "ArtifactCollectionScope",
    "ProjectScope",
]



__all__ = (
    "Error",
    "CommError",
    "AuthenticationError",
    "UsageError",
    "UnsupportedError",
    "WandbCoreNotAvailableError",
)

from .errors import (
    AuthenticationError,
    CommError,
    Error,
    UnsupportedError,
    UsageError,
    WandbCoreNotAvailableError,
)



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/projects/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase


class RenameProject(GQLBase):
    rename_project: Optional[RenameProjectRenameProject] = Field(alias="renameProject")


class RenameProjectRenameProject(GQLBase):
    project: Optional[RenameProjectRenameProjectProject]
    inserted: Optional[bool]


class RenameProjectRenameProjectProject(GQLBase):
    name: str


RenameProject.model_rebuild()
RenameProjectRenameProject.model_rebuild()



from __future__ import annotations

from enum import Enum
from functools import singledispatch
from itertools import chain
from typing import Any, TypeVar

from pydantic_core import PydanticUseDefault

from ._filters import And, FilterExpr, In, Nor, Not, NotIn, Op, Or

T = TypeVar("T")


class LenientStrEnum(str, Enum):
    """A string enum allowing for case-insensitive lookups by value.

    May include other internal customizations if needed.

    Note: This is a bespoke, internal implementation and NOT intended as a
    backport of `enum.StrEnum` from Python 3.11+.
    """

    def __repr__(self) -> str:
        return self.name

    @classmethod
    def _missing_(cls, value: object) -> Any:
        # Accept case-insensitive enum values
        if isinstance(value, str):
            v = value.lower()
            return next((e for e in cls if e.value.lower() == v), None)
        return None


def default_if_none(v: Any) -> Any:
    """A before-validator validator that coerces `None` to the default field value instead."""
    # https://docs.pydantic.dev/2.11/api/pydantic_core/#pydantic_core.PydanticUseDefault
    if v is None:
        raise PydanticUseDefault
    return v


def upper_if_str(v: Any) -> Any:
    return v.strip().upper() if isinstance(v, str) else v


# ----------------------------------------------------------------------------
def to_scope(v: Any) -> Any:
    """Convert eligible objects, including pre-existing `wandb` types, to an automation scope."""
    from wandb.apis.public import ArtifactCollection, Project

    from .scopes import ProjectScope, _ArtifactPortfolioScope, _ArtifactSequenceScope

    if isinstance(v, Project):
        return ProjectScope(id=v.id, name=v.name)
    if isinstance(v, ArtifactCollection):
        cls = _ArtifactSequenceScope if v.is_sequence() else _ArtifactPortfolioScope
        return cls(id=v.id, name=v.name)
    return v


def to_saved_action(v: Any) -> Any:
    """If necessary (and possible), convert the object to a saved action."""
    from .actions import (
        DoNothing,
        SavedNoOpAction,
        SavedNotificationAction,
        SavedWebhookAction,
        SendNotification,
        SendWebhook,
    )

    if isinstance(v, SendNotification):
        return SavedNotificationAction(
            integration={"id": v.integration_id},
            **v.model_dump(exclude={"integration_id"}),
        )
    if isinstance(v, SendWebhook):
        return SavedWebhookAction(
            integration={"id": v.integration_id},
            **v.model_dump(exclude={"integration_id"}),
        )
    if isinstance(v, DoNothing):
        return SavedNoOpAction.model_validate(v)

    return v


def to_input_action(v: Any) -> Any:
    """If necessary (and possible), convert the object to an input action."""
    from .actions import (
        DoNothing,
        SavedNoOpAction,
        SavedNotificationAction,
        SavedWebhookAction,
        SendNotification,
        SendWebhook,
    )

    if isinstance(v, SavedNotificationAction):
        return SendNotification(
            integration_id=v.integration.id,
            **v.model_dump(exclude={"integration"}),
        )
    if isinstance(v, SavedWebhookAction):
        return SendWebhook(
            integration_id=v.integration.id,
            **v.model_dump(exclude={"integration"}),
        )
    if isinstance(v, SavedNoOpAction):
        return DoNothing.model_validate(v)

    return v


# ----------------------------------------------------------------------------
@singledispatch
def simplify_op(op: Op | FilterExpr) -> Op | FilterExpr:
    """Simplify a MongoDB filter by removing and unnesting redundant operators."""
    return op


@simplify_op.register
def _(op: And) -> Op:
    # {"$and": []} -> {"$and": []}
    if not (args := op.and_):
        return op

    # {"$and": [op]} -> op
    if len(args) == 1:
        return simplify_op(args[0])

    # {"$and": [op, {"$and": [op2, ...]}]} -> {"$and": [op, op2, ...]}
    flattened = chain.from_iterable(x.and_ if isinstance(x, And) else [x] for x in args)
    return And(and_=map(simplify_op, flattened))


@simplify_op.register
def _(op: Or) -> Op:
    # {"$or": []} -> {"$or": []}
    if not (args := op.or_):
        return op

    # {"$or": [op]} -> op
    if len(args) == 1:
        return simplify_op(args[0])

    # {"$or": [op, {"$or": [op2, ...]}]} -> {"$or": [op, op2, ...]}
    flattened = chain.from_iterable(x.or_ if isinstance(x, Or) else [x] for x in args)
    return Or(or_=map(simplify_op, flattened))


@simplify_op.register
def _(op: Not) -> Op:
    inner = op.not_

    # {"$not": {"$not": op}} -> op
    # {"$not": {"$or": [op, ...]}} -> {"$nor": [op, ...]}
    # {"$not": {"$nor": [op, ...]}} -> {"$or": [op, ...]}
    # {"$not": {"$in": [op, ...]}} -> {"$nin": [op, ...]}
    # {"$not": {"$nin": [op, ...]}} -> {"$in": [op, ...]}
    if isinstance(inner, (Not, Or, Nor, In, NotIn)):
        return simplify_op(~inner)
    return Not(not_=simplify_op(inner))



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/projects/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import RegistryFragment


class UpsertRegistryProject(GQLBase):
    upsert_model: Optional[UpsertRegistryProjectUpsertModel] = Field(
        alias="upsertModel"
    )


class UpsertRegistryProjectUpsertModel(GQLBase):
    project: Optional[RegistryFragment]
    inserted: Optional[bool]


UpsertRegistryProject.model_rebuild()
UpsertRegistryProjectUpsertModel.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Literal, Optional

from pydantic import Field

from wandb._pydantic import GQLBase, Typename

from .fragments import FilesFragment


class ArtifactCollectionMembershipFiles(GQLBase):
    project: Optional[ArtifactCollectionMembershipFilesProject]


class ArtifactCollectionMembershipFilesProject(GQLBase):
    artifact_collection: Optional[
        ArtifactCollectionMembershipFilesProjectArtifactCollection
    ] = Field(alias="artifactCollection")


class ArtifactCollectionMembershipFilesProjectArtifactCollection(GQLBase):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    artifact_membership: Optional[
        ArtifactCollectionMembershipFilesProjectArtifactCollectionArtifactMembership
    ] = Field(alias="artifactMembership")


class ArtifactCollectionMembershipFilesProjectArtifactCollectionArtifactMembership(
    GQLBase
):
    files: Optional[FilesFragment]


ArtifactCollectionMembershipFiles.model_rebuild()
ArtifactCollectionMembershipFilesProject.model_rebuild()
ArtifactCollectionMembershipFilesProjectArtifactCollection.model_rebuild()
ArtifactCollectionMembershipFilesProjectArtifactCollectionArtifactMembership.model_rebuild()



# Generated by ariadne-codegen
# Source: core/api/graphql/schemas/schema-latest.graphql

from __future__ import annotations

from enum import Enum


class EventTriggeringConditionType(str, Enum):
    CREATE_ARTIFACT = "CREATE_ARTIFACT"
    UPDATE_ARTIFACT_ALIAS = "UPDATE_ARTIFACT_ALIAS"
    ADD_ARTIFACT_ALIAS = "ADD_ARTIFACT_ALIAS"
    LINK_MODEL = "LINK_MODEL"
    RUN_METRIC = "RUN_METRIC"
    RUN_METRIC_CHANGE = "RUN_METRIC_CHANGE"


class TriggerScopeType(str, Enum):
    PROJECT = "PROJECT"
    ARTIFACT_COLLECTION = "ARTIFACT_COLLECTION"


class AlertSeverity(str, Enum):
    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"


class TriggeredActionType(str, Enum):
    QUEUE_JOB = "QUEUE_JOB"
    NOTIFICATION = "NOTIFICATION"
    GENERIC_WEBHOOK = "GENERIC_WEBHOOK"
    NO_OP = "NO_OP"



# Generated by ariadne-codegen

from .create_automation import CreateAutomation
from .create_generic_webhook_integration import (
    CreateGenericWebhookIntegration,
    CreateGenericWebhookIntegrationCreateGenericWebhookIntegration,
    CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationGenericWebhookIntegration,
    CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationIntegration,
)
from .delete_automation import DeleteAutomation
from .enums import (
    AlertSeverity,
    EventTriggeringConditionType,
    TriggeredActionType,
    TriggerScopeType,
)
from .fragments import (
    ArtifactPortfolioScopeFields,
    ArtifactSequenceScopeFields,
    FilterEventFields,
    GenericWebhookActionFields,
    GenericWebhookActionFieldsIntegrationGenericWebhookIntegration,
    GenericWebhookActionFieldsIntegrationIntegration,
    GenericWebhookIntegrationConnectionFields,
    GenericWebhookIntegrationConnectionFieldsEdges,
    GenericWebhookIntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration,
    GenericWebhookIntegrationConnectionFieldsEdgesNodeIntegration,
    GenericWebhookIntegrationFields,
    IntegrationConnectionFields,
    IntegrationConnectionFieldsEdges,
    IntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration,
    IntegrationConnectionFieldsEdgesNodeIntegration,
    IntegrationConnectionFieldsEdgesNodeSlackIntegration,
    NoOpActionFields,
    NotificationActionFields,
    NotificationActionFieldsIntegrationIntegration,
    NotificationActionFieldsIntegrationSlackIntegration,
    PageInfoFields,
    ProjectConnectionFields,
    ProjectConnectionFieldsEdges,
    ProjectConnectionFieldsEdgesNode,
    ProjectScopeFields,
    QueueJobActionFields,
    QueueJobActionFieldsQueue,
    SlackIntegrationConnectionFields,
    SlackIntegrationConnectionFieldsEdges,
    SlackIntegrationConnectionFieldsEdgesNodeIntegration,
    SlackIntegrationConnectionFieldsEdgesNodeSlackIntegration,
    SlackIntegrationFields,
    TriggerFields,
    TriggerFieldsActionGenericWebhookTriggeredAction,
    TriggerFieldsActionNoOpTriggeredAction,
    TriggerFieldsActionNotificationTriggeredAction,
    TriggerFieldsActionQueueJobTriggeredAction,
    TriggerFieldsEventFilterEventTriggeringCondition,
    TriggerFieldsScopeArtifactPortfolio,
    TriggerFieldsScopeArtifactSequence,
    TriggerFieldsScopeProject,
)
from .generic_webhook_integrations_by_entity import (
    GenericWebhookIntegrationsByEntity,
    GenericWebhookIntegrationsByEntityEntity,
)
from .get_automations import GetAutomations, GetAutomationsSearchScope
from .get_automations_by_entity import (
    GetAutomationsByEntity,
    GetAutomationsByEntitySearchScope,
)
from .input_types import (
    CreateFilterTriggerInput,
    CreateGenericWebhookIntegrationInput,
    GenericWebhookActionInput,
    NoOpTriggeredActionInput,
    NotificationActionInput,
    QueueJobActionInput,
    TriggeredActionConfig,
    UpdateFilterTriggerInput,
)
from .integrations_by_entity import IntegrationsByEntity, IntegrationsByEntityEntity
from .operations import (
    CREATE_AUTOMATION_GQL,
    CREATE_GENERIC_WEBHOOK_INTEGRATION_GQL,
    DELETE_AUTOMATION_GQL,
    GENERIC_WEBHOOK_INTEGRATIONS_BY_ENTITY_GQL,
    GET_AUTOMATIONS_BY_ENTITY_GQL,
    GET_AUTOMATIONS_GQL,
    INTEGRATIONS_BY_ENTITY_GQL,
    SLACK_INTEGRATIONS_BY_ENTITY_GQL,
    UPDATE_AUTOMATION_GQL,
)
from .slack_integrations_by_entity import (
    SlackIntegrationsByEntity,
    SlackIntegrationsByEntityEntity,
)
from .update_automation import UpdateAutomation

__all__ = [
    "CREATE_AUTOMATION_GQL",
    "CREATE_GENERIC_WEBHOOK_INTEGRATION_GQL",
    "DELETE_AUTOMATION_GQL",
    "GENERIC_WEBHOOK_INTEGRATIONS_BY_ENTITY_GQL",
    "GET_AUTOMATIONS_BY_ENTITY_GQL",
    "GET_AUTOMATIONS_GQL",
    "INTEGRATIONS_BY_ENTITY_GQL",
    "SLACK_INTEGRATIONS_BY_ENTITY_GQL",
    "UPDATE_AUTOMATION_GQL",
    "GetAutomations",
    "GetAutomationsSearchScope",
    "GetAutomationsByEntity",
    "GetAutomationsByEntitySearchScope",
    "CreateAutomation",
    "UpdateAutomation",
    "DeleteAutomation",
    "IntegrationsByEntity",
    "IntegrationsByEntityEntity",
    "SlackIntegrationsByEntity",
    "SlackIntegrationsByEntityEntity",
    "GenericWebhookIntegrationsByEntity",
    "GenericWebhookIntegrationsByEntityEntity",
    "CreateGenericWebhookIntegration",
    "CreateGenericWebhookIntegrationCreateGenericWebhookIntegration",
    "CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationGenericWebhookIntegration",
    "CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationIntegration",
    "CreateFilterTriggerInput",
    "CreateGenericWebhookIntegrationInput",
    "GenericWebhookActionInput",
    "NoOpTriggeredActionInput",
    "NotificationActionInput",
    "QueueJobActionInput",
    "TriggeredActionConfig",
    "UpdateFilterTriggerInput",
    "ArtifactPortfolioScopeFields",
    "ArtifactSequenceScopeFields",
    "FilterEventFields",
    "GenericWebhookActionFields",
    "GenericWebhookActionFieldsIntegrationGenericWebhookIntegration",
    "GenericWebhookActionFieldsIntegrationIntegration",
    "GenericWebhookIntegrationConnectionFields",
    "GenericWebhookIntegrationConnectionFieldsEdges",
    "GenericWebhookIntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration",
    "GenericWebhookIntegrationConnectionFieldsEdgesNodeIntegration",
    "GenericWebhookIntegrationFields",
    "IntegrationConnectionFields",
    "IntegrationConnectionFieldsEdges",
    "IntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration",
    "IntegrationConnectionFieldsEdgesNodeIntegration",
    "IntegrationConnectionFieldsEdgesNodeSlackIntegration",
    "NoOpActionFields",
    "NotificationActionFields",
    "NotificationActionFieldsIntegrationIntegration",
    "NotificationActionFieldsIntegrationSlackIntegration",
    "PageInfoFields",
    "ProjectConnectionFields",
    "ProjectConnectionFieldsEdges",
    "ProjectConnectionFieldsEdgesNode",
    "ProjectScopeFields",
    "QueueJobActionFields",
    "QueueJobActionFieldsQueue",
    "SlackIntegrationConnectionFields",
    "SlackIntegrationConnectionFieldsEdges",
    "SlackIntegrationConnectionFieldsEdgesNodeIntegration",
    "SlackIntegrationConnectionFieldsEdgesNodeSlackIntegration",
    "SlackIntegrationFields",
    "TriggerFields",
    "TriggerFieldsActionGenericWebhookTriggeredAction",
    "TriggerFieldsActionNoOpTriggeredAction",
    "TriggerFieldsActionNotificationTriggeredAction",
    "TriggerFieldsActionQueueJobTriggeredAction",
    "TriggerFieldsEventFilterEventTriggeringCondition",
    "TriggerFieldsScopeArtifactPortfolio",
    "TriggerFieldsScopeArtifactSequence",
    "TriggerFieldsScopeProject",
    "AlertSeverity",
    "EventTriggeringConditionType",
    "TriggerScopeType",
    "TriggeredActionType",
]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Literal, Optional, Union

from pydantic import Field

from wandb._pydantic import GQLBase, Typename

from .fragments import GenericWebhookIntegrationFields


class CreateGenericWebhookIntegration(GQLBase):
    create_generic_webhook_integration: Optional[
        CreateGenericWebhookIntegrationCreateGenericWebhookIntegration
    ] = Field(alias="createGenericWebhookIntegration")


class CreateGenericWebhookIntegrationCreateGenericWebhookIntegration(GQLBase):
    integration: Union[
        CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationIntegration,
        CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationGenericWebhookIntegration,
    ] = Field(discriminator="typename__")


class CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationIntegration(
    GQLBase
):
    typename__: Typename[
        Literal["GitHubOAuthIntegration", "Integration", "SlackIntegration"]
    ]


class CreateGenericWebhookIntegrationCreateGenericWebhookIntegrationIntegrationGenericWebhookIntegration(
    GenericWebhookIntegrationFields
):
    typename__: Typename[Literal["GenericWebhookIntegration"]]


CreateGenericWebhookIntegration.model_rebuild()
CreateGenericWebhookIntegrationCreateGenericWebhookIntegration.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase


class AddAliases(GQLBase):
    add_aliases: Optional[AddAliasesAddAliases] = Field(alias="addAliases")


class AddAliasesAddAliases(GQLBase):
    success: bool


AddAliases.model_rebuild()



# Generated by ariadne-codegen

from .operations import SERVER_FEATURES_QUERY_GQL
from .server_features_query import (
    ServerFeaturesQuery,
    ServerFeaturesQueryServerInfo,
    ServerFeaturesQueryServerInfoFeatures,
)

__all__ = [
    "SERVER_FEATURES_QUERY_GQL",
    "ServerFeaturesQuery",
    "ServerFeaturesQueryServerInfo",
    "ServerFeaturesQueryServerInfoFeatures",
]



"""Internal utilities for working with pydantic."""

from .base import (
    CompatBaseModel,
    GQLBase,
    GQLId,
    SerializedToJson,
    Typename,
    ensure_json,
)
from .utils import IS_PYDANTIC_V2, from_json, pydantic_isinstance, to_json
from .v1_compat import AliasChoices, computed_field, field_validator, model_validator

__all__ = [
    "IS_PYDANTIC_V2",
    "CompatBaseModel",
    "GQLBase",
    "Typename",
    "GQLId",
    "SerializedToJson",
    "AliasChoices",
    "computed_field",
    "field_validator",
    "model_validator",
    "pydantic_isinstance",
    "to_json",
    "from_json",
    "ensure_json",
]



from enum import Enum
from functools import lru_cache
from typing import TYPE_CHECKING, Any, List, Literal, Mapping, Optional, Sequence

from wandb.sdk.artifacts._validators import (
    REGISTRY_PREFIX,
    validate_artifact_types_list,
)

if TYPE_CHECKING:
    from wandb_gql import Client

from wandb_gql import gql


class _Visibility(str, Enum):
    # names are what users see/pass into Python methods
    # values are what's expected by backend API
    organization = "PRIVATE"
    restricted = "RESTRICTED"

    @classmethod
    def _missing_(cls, value: object) -> Any:
        return next(
            (e for e in cls if e.name == value),
            None,
        )


def _format_gql_artifact_types_input(
    artifact_types: Optional[List[str]] = None,
):
    """Format the artifact types for the GQL input.

    Args:
        artifact_types: The artifact types to add to the registry.

    Returns:
        The artifact types for the GQL input.
    """
    if artifact_types is None:
        return []
    new_types = validate_artifact_types_list(artifact_types)
    return [{"name": type} for type in new_types]


def _gql_to_registry_visibility(
    visibility: str,
) -> Literal["organization", "restricted"]:
    """Convert the GQL visibility to the registry visibility.

    Args:
        visibility: The GQL visibility.

    Returns:
        The registry visibility.
    """
    try:
        return _Visibility(visibility).name
    except ValueError:
        raise ValueError(f"Invalid visibility: {visibility!r} from backend")


def _registry_visibility_to_gql(
    visibility: Literal["organization", "restricted"],
) -> str:
    """Convert the registry visibility to the GQL visibility."""
    try:
        return _Visibility[visibility].value
    except KeyError:
        raise ValueError(
            f"Invalid visibility: {visibility!r}. "
            f"Must be one of: {', '.join(map(repr, (e.name for e in _Visibility)))}"
        )


def _ensure_registry_prefix_on_names(query, in_name=False):
    """Traverse the filter to prepend the `name` key value with the registry prefix unless the value is a regex.

    - in_name: True if we are under a "name" key (or propagating from one).

    EX: {"name": "model"} -> {"name": "wandb-registry-model"}
    """
    if isinstance((txt := query), str):
        if in_name:
            return txt if txt.startswith(REGISTRY_PREFIX) else f"{REGISTRY_PREFIX}{txt}"
        return txt
    if isinstance((dct := query), Mapping):
        new_dict = {}
        for key, obj in dct.items():
            if key == "name":
                new_dict[key] = _ensure_registry_prefix_on_names(obj, in_name=True)
            elif key == "$regex":
                # For regex operator, we skip transformation of its value.
                new_dict[key] = obj
            else:
                # For any other key, propagate the in_name and skip_transform flags as-is.
                new_dict[key] = _ensure_registry_prefix_on_names(obj, in_name=in_name)
        return new_dict
    if isinstance((objs := query), Sequence):
        return list(
            map(lambda x: _ensure_registry_prefix_on_names(x, in_name=in_name), objs)
        )
    return query


@lru_cache(maxsize=10)
def _fetch_org_entity_from_organization(client: "Client", organization: str) -> str:
    """Fetch the org entity from the organization.

    Args:
        client (Client): Graphql client.
        organization (str): The organization to fetch the org entity for.
    """
    query = gql(
        """
        query FetchOrgEntityFromOrganization($organization: String!) {
            organization(name: $organization) {
                    orgEntity {
                        name
                    }
                }
            }
        """
    )
    try:
        response = client.execute(query, variable_values={"organization": organization})
        if response["organization"] and response["organization"]["orgEntity"]:
            if response["organization"]["orgEntity"]["name"]:
                return response["organization"]["orgEntity"]["name"]
            return ValueError(
                f"Organization entity for organization: {organization} is empty"
            )
        raise ValueError(
            f"Organization entity for organization: {organization} not found"
        )
    except Exception as e:
        raise ValueError(
            f"Error fetching org entity for organization: {organization}"
        ) from e



#
"""Setup wandb session.

This module configures a wandb session which can extend to multiple wandb runs.

Functions:
    setup(): Configure wandb session.

Early logging keeps track of logger output until the call to wandb.init() when the
run_id can be resolved.

"""

from __future__ import annotations

import logging
import os
import pathlib
import sys
import threading
from typing import TYPE_CHECKING, Any, Union

import wandb
import wandb.integration.sagemaker as sagemaker
from wandb.env import CONFIG_DIR
from wandb.sdk.lib import import_hooks, wb_logging

from . import wandb_settings
from .lib import config_util, server

if TYPE_CHECKING:
    from wandb.sdk import wandb_run
    from wandb.sdk.lib.service_connection import ServiceConnection
    from wandb.sdk.wandb_settings import Settings


class _EarlyLogger:
    """Early logger which captures logs in memory until logging can be configured."""

    def __init__(self) -> None:
        self._log: list[tuple] = []
        self._exception: list[tuple] = []
        # support old warn() as alias of warning()
        self.warn = self.warning

    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.DEBUG, msg, args, kwargs))

    def info(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.INFO, msg, args, kwargs))

    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.WARNING, msg, args, kwargs))

    def error(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.ERROR, msg, args, kwargs))

    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.CRITICAL, msg, args, kwargs))

    def exception(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._exception.append((msg, args, kwargs))

    def log(self, level: str, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((level, msg, args, kwargs))

    def _flush(self, new_logger: Logger) -> None:
        assert self is not new_logger
        for level, msg, args, kwargs in self._log:
            new_logger.log(level, msg, *args, **kwargs)
        for msg, args, kwargs in self._exception:
            new_logger.exception(msg, *args, **kwargs)


Logger = Union[logging.Logger, _EarlyLogger]


class _WandbSetup:
    """W&B library singleton."""

    def __init__(
        self,
        pid: int,
        settings: Settings | None = None,
        environ: dict | None = None,
    ) -> None:
        self._connection: ServiceConnection | None = None

        self._active_runs: list[wandb_run.Run] = []

        self._environ = environ or dict(os.environ)
        self._sweep_config: dict | None = None
        self._config: dict | None = None
        self._server: server.Server | None = None
        self._pid = pid

        # TODO(jhr): defer strict checks until settings are fully initialized
        #            and logging is ready
        self._logger: Logger = _EarlyLogger()

        self._settings = self._settings_setup(settings)

        wandb.termsetup(self._settings, None)

        self._check()
        self._setup()

    def add_active_run(self, run: wandb_run.Run) -> None:
        """Append a run to the active runs list.

        This must be called when a run is initialized.

        Args:
            run: A newly initialized run.
        """
        if run not in self._active_runs:
            self._active_runs.append(run)

    def remove_active_run(self, run: wandb_run.Run) -> None:
        """Remove the run from the active runs list.

        This must be called when a run is finished.

        Args:
            run: A run that is finished or crashed.
        """
        try:
            self._active_runs.remove(run)
        except ValueError:
            pass  # Removing a run multiple times is not an error.

    @property
    def most_recent_active_run(self) -> wandb_run.Run | None:
        """The most recently initialized run that is not yet finished."""
        if not self._active_runs:
            return None

        return self._active_runs[-1]

    def finish_all_active_runs(self) -> None:
        """Finish all unfinished runs.

        NOTE: This is slightly inefficient as it finishes runs one at a time.
        This only exists to support using the `reinit="finish_previous"`
        setting together with `reinit="create_new"` which does not seem to be a
        useful pattern. Since `"create_new"` should eventually become the
        default and only behavior, it does not seem worth optimizing.
        """
        # Take a snapshot as each call to `finish()` modifies `_active_runs`.
        runs_copy = list(self._active_runs)
        for run in runs_copy:
            run.finish()

    def _settings_setup(
        self,
        settings: Settings | None,
    ) -> wandb_settings.Settings:
        s = wandb_settings.Settings()

        # the pid of the process to monitor for system stats
        pid = os.getpid()
        self._logger.info(f"Current SDK version is {wandb.__version__}")
        self._logger.info(f"Configure stats pid to {pid}")
        s.x_stats_pid = pid

        if settings and settings.settings_system:
            s.settings_system = settings.settings_system
        elif config_dir_str := os.getenv(CONFIG_DIR, None):
            config_dir = pathlib.Path(config_dir_str).expanduser()
            s.settings_system = str(config_dir / "settings")
        else:
            s.settings_system = str(
                pathlib.Path("~", ".config", "wandb", "settings").expanduser()
            )

        # load settings from the system config
        if s.settings_system:
            self._logger.info(f"Loading settings from {s.settings_system}")
        s.update_from_system_config_file()

        # load settings from the workspace config
        if s.settings_workspace:
            self._logger.info(f"Loading settings from {s.settings_workspace}")
        s.update_from_workspace_config_file()

        # load settings from the environment variables
        self._logger.info("Loading settings from environment variables")
        s.update_from_env_vars(self._environ)

        # infer settings from the system environment
        s.update_from_system_environment()

        # load SageMaker settings
        check_sagemaker_env = not s.sagemaker_disable
        if settings and settings.sagemaker_disable:
            check_sagemaker_env = False
        if check_sagemaker_env and sagemaker.is_using_sagemaker():
            self._logger.info("Loading SageMaker settings")
            sagemaker.set_global_settings(s)

        # load settings from the passed init/setup settings
        if settings:
            s.update_from_settings(settings)

        return s

    def _update(self, settings: Settings | None = None) -> None:
        if not settings:
            return
        self._settings.update_from_settings(settings)

    def _update_user_settings(self) -> None:
        # Get rid of cached results to force a refresh.
        self._server = None
        user_settings = self._load_user_settings()
        if user_settings is not None:
            self._settings.update_from_dict(user_settings)

    def _early_logger_flush(self, new_logger: Logger) -> None:
        if self._logger is new_logger:
            return

        if isinstance(self._logger, _EarlyLogger):
            self._logger._flush(new_logger)
        self._logger = new_logger

    def _get_logger(self) -> Logger:
        return self._logger

    @property
    def settings(self) -> wandb_settings.Settings:
        return self._settings

    def _get_entity(self) -> str | None:
        if self._settings and self._settings._offline:
            return None
        entity = self.viewer.get("entity")
        return entity

    def _get_username(self) -> str | None:
        if self._settings and self._settings._offline:
            return None
        return self.viewer.get("username")

    def _get_teams(self) -> list[str]:
        if self._settings and self._settings._offline:
            return []
        teams = self.viewer.get("teams")
        if teams:
            teams = [team["node"]["name"] for team in teams["edges"]]
        return teams or []

    @property
    def viewer(self) -> dict[str, Any]:
        if self._server is None:
            self._server = server.Server(settings=self._settings)

        return self._server.viewer

    def _load_user_settings(self) -> dict[str, Any] | None:
        # offline?
        if self._server is None:
            return None

        flags = self._server._flags
        user_settings = dict()
        if "code_saving_enabled" in flags:
            user_settings["save_code"] = flags["code_saving_enabled"]

        email = self.viewer.get("email", None)
        if email:
            user_settings["email"] = email

        return user_settings

    def _check(self) -> None:
        if hasattr(threading, "main_thread"):
            if threading.current_thread() is not threading.main_thread():
                pass
        elif threading.current_thread().name != "MainThread":
            wandb.termwarn(f"bad thread2: {threading.current_thread().name}")
        if getattr(sys, "frozen", False):
            wandb.termwarn("frozen, could be trouble")

    def _setup(self) -> None:
        sweep_path = self._settings.sweep_param_path
        if sweep_path:
            self._sweep_config = config_util.dict_from_config_file(
                sweep_path, must_exist=True
            )

        # if config_paths was set, read in config dict
        if self._settings.config_paths:
            # TODO(jhr): handle load errors, handle list of files
            for config_path in self._settings.config_paths:
                config_dict = config_util.dict_from_config_file(config_path)
                if config_dict is None:
                    continue
                if self._config is not None:
                    self._config.update(config_dict)
                else:
                    self._config = config_dict

    def _teardown(self, exit_code: int | None = None) -> None:
        import_hooks.unregister_all_post_import_hooks()

        if not self._connection:
            return

        internal_exit_code = self._connection.teardown(exit_code or 0)

        # Reset to None so that setup() creates a new connection.
        self._connection = None

        if internal_exit_code != 0:
            sys.exit(internal_exit_code)

    def ensure_service(self) -> ServiceConnection:
        """Returns a connection to the service process creating it if needed."""
        if self._connection:
            return self._connection

        from wandb.sdk.lib import service_connection

        self._connection = service_connection.connect_to_service(self._settings)
        return self._connection

    def assert_service(self) -> ServiceConnection:
        """Returns a connection to the service process, asserting it exists.

        Unlike ensure_service(), this will not start up a service process
        if it didn't already exist.
        """
        if not self._connection:
            raise AssertionError("Expected service process to exist.")

        return self._connection


_singleton: _WandbSetup | None = None
"""The W&B library singleton, or None if not yet set up.

The value is invalid and must not be used if `os.getpid() != _singleton._pid`.
"""


def singleton() -> _WandbSetup | None:
    """Returns the W&B singleton if it exists for the current process.

    Unlike setup(), this does not create the singleton if it doesn't exist.
    """
    if _singleton and _singleton._pid == os.getpid():
        return _singleton
    else:
        return None


@wb_logging.log_to_all_runs()
def _setup(
    settings: Settings | None = None,
    start_service: bool = True,
) -> _WandbSetup:
    """Set up library context.

    Args:
        settings: Global settings to set, or updates to the global settings
            if the singleton has already been initialized.
        start_service: Whether to start up the service process.
            NOTE: A service process will only be started if allowed by the
            global settings (after the given updates). The service will not
            start up if the mode resolves to "disabled".
    """
    global _singleton

    pid = os.getpid()

    if _singleton and _singleton._pid == pid:
        _singleton._update(settings=settings)
    else:
        _singleton = _WandbSetup(settings=settings, pid=pid)

    if start_service and not _singleton.settings._noop:
        _singleton.ensure_service()

    return _singleton


def setup(settings: Settings | None = None) -> _WandbSetup:
    """Prepares W&B for use in the current process and its children.

    You can usually ignore this as it is implicitly called by `wandb.init()`.

    When using wandb in multiple processes, calling `wandb.setup()`
    in the parent process before starting child processes may improve
    performance and resource utilization.

    Note that `wandb.setup()` modifies `os.environ`, and it is important
    that child processes inherit the modified environment variables.

    See also `wandb.teardown()`.

    Args:
        settings: Configuration settings to apply globally. These can be
            overridden by subsequent `wandb.init()` calls.

    Example:
        ```python
        import multiprocessing

        import wandb


        def run_experiment(params):
            with wandb.init(config=params):
                # Run experiment
                pass


        if __name__ == "__main__":
            # Start backend and set global config
            wandb.setup(settings={"project": "my_project"})

            # Define experiment parameters
            experiment_params = [
                {"learning_rate": 0.01, "epochs": 10},
                {"learning_rate": 0.001, "epochs": 20},
            ]

            # Start multiple processes, each running a separate experiment
            processes = []
            for params in experiment_params:
                p = multiprocessing.Process(target=run_experiment, args=(params,))
                p.start()
                processes.append(p)

            # Wait for all processes to complete
            for p in processes:
                p.join()

            # Optional: Explicitly shut down the backend
            wandb.teardown()
        ```
    """
    return _setup(settings=settings)


@wb_logging.log_to_all_runs()
def teardown(exit_code: int | None = None) -> None:
    """Waits for wandb to finish and frees resources.

    Completes any runs that were not explicitly finished
    using `run.finish()` and waits for all data to be uploaded.

    It is recommended to call this at the end of a session
    that used `wandb.setup()`. It is invoked automatically
    in an `atexit` hook, but this is not reliable in certain setups
    such as when using Python's `multiprocessing` module.
    """
    global _singleton

    orig_singleton = _singleton
    _singleton = None

    if orig_singleton:
        orig_singleton._teardown(exit_code=exit_code)



from __future__ import annotations

import dataclasses
import os

from wandb import env

_CURRENT_VERSION = "2"
_SUPPORTED_TRANSPORTS = "tcp"


def get_service_token() -> ServiceToken | None:
    """Reads the token from environment variables.

    Returns:
        The token if the correct environment variable is set, or None.

    Raises:
        ValueError: If the environment variable is set but cannot be
            parsed.
    """
    token = os.environ.get(env.SERVICE)
    if not token:
        return None

    parts = token.split("-")
    if len(parts) != 5:
        raise ValueError(f"Invalid token: {token}")

    version, pid_str, transport, host, port_str = parts

    if version != _CURRENT_VERSION:
        raise ValueError(
            f"Expected version {_CURRENT_VERSION},"
            f" but got {version} (token={token})"
        )
    if transport not in _SUPPORTED_TRANSPORTS:
        raise ValueError(
            f"Unsupported transport: {transport} (token={token})",
        )

    try:
        return ServiceToken(
            version=version,
            pid=int(pid_str),
            transport=transport,
            host=host,
            port=int(port_str),
        )
    except ValueError as e:
        raise ValueError(f"Invalid token: {token}") from e


def set_service_token(parent_pid: int, transport: str, host: str, port: int) -> None:
    """Stores a service token in an environment variable.

    Args:
        parent_pid: The process ID of the process that started the service.
        transport: The transport used to communicate with the service.
        host: The host part of the internet address on which the service
            is listening (e.g. localhost).
        port: The port the service is listening on.

    Raises:
        ValueError: If given an unsupported transport.
    """
    if transport not in _SUPPORTED_TRANSPORTS:
        raise ValueError(f"Unsupported transport: {transport}")

    os.environ[env.SERVICE] = "-".join(
        (
            _CURRENT_VERSION,
            str(parent_pid),
            transport,
            host,
            str(port),
        )
    )


def clear_service_token() -> None:
    """Clears the environment variable storing the service token."""
    os.environ.pop(env.SERVICE, None)


@dataclasses.dataclass(frozen=True)
class ServiceToken:
    """An identifier for a running service process."""

    version: str
    pid: int
    transport: str
    host: str
    port: int



"""Functions for compatibility with asyncio."""

from __future__ import annotations

import asyncio
import concurrent
import concurrent.futures
import contextlib
import threading
from typing import Any, AsyncIterator, Callable, Coroutine, Iterator, TypeVar

_T = TypeVar("_T")


def run(fn: Callable[[], Coroutine[Any, Any, _T]]) -> _T:
    """Run `fn` in an asyncio loop in a new thread.

    This must always be used instead of `asyncio.run` which fails if there is
    an active `asyncio` event loop in the current thread. Since `wandb` was not
    originally designed with `asyncio` in mind, using `asyncio.run` would break
    users who were calling `wandb` methods from an `asyncio` loop.

    Note that due to starting a new thread, this is slightly slow.
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        runner = _Runner()
        future = executor.submit(runner.run, fn)

        try:
            return future.result()

        finally:
            runner.cancel()


class _RunnerCancelledError(Exception):
    """The `_Runner.run()` invocation was cancelled."""


class _Runner:
    """Runs an asyncio event loop allowing cancellation.

    This is like `asyncio.run()`, except it provides a `cancel()` method
    meant to be called in a `finally` block.

    Without this, it is impossible to make `asyncio.run()` stop if it runs
    in a non-main thread. In particular, a KeyboardInterrupt causes the
    ThreadPoolExecutor above to block until the asyncio thread completes,
    but there is no way to tell the asyncio thread to cancel its work.
    A second KeyboardInterrupt makes ThreadPoolExecutor give up while the
    asyncio thread still runs in the background, with terrible effects if it
    prints to the user's terminal.
    """

    def __init__(self) -> None:
        self._lock = threading.Lock()

        self._is_cancelled = False
        self._started = False
        self._done = False

        self._loop: asyncio.AbstractEventLoop | None = None
        self._cancel_event: asyncio.Event | None = None

    def run(self, fn: Callable[[], Coroutine[Any, Any, _T]]) -> _T:
        """Run a coroutine in asyncio, cancelling it on `cancel()`.

        Returns:
            The result of the coroutine returned by `fn`.

        Raises:
            _RunnerCancelledError: If `cancel()` is called.
        """
        return asyncio.run(self._run_or_cancel(fn))

    async def _run_or_cancel(
        self,
        fn: Callable[[], Coroutine[Any, Any, _T]],
    ) -> _T:
        with self._lock:
            if self._is_cancelled:
                raise _RunnerCancelledError()

            self._loop = asyncio.get_running_loop()
            self._cancel_event = asyncio.Event()
            self._started = True

        cancellation_task = asyncio.create_task(self._cancel_event.wait())
        fn_task = asyncio.create_task(fn())

        try:
            await asyncio.wait(
                [cancellation_task, fn_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if fn_task.done():
                return fn_task.result()
            else:
                raise _RunnerCancelledError()

        finally:
            cancellation_task.cancel()
            fn_task.cancel()

            with self._lock:
                self._done = True

    def cancel(self) -> None:
        """Cancel all asyncio work started by `run()`."""
        with self._lock:
            if self._is_cancelled:
                return
            self._is_cancelled = True

            if self._done or not self._started:
                # If the runner already finished, no need to cancel it.
                #
                # If the runner hasn't started the loop yet, then it will not
                # as we already set _is_cancelled.
                return

            assert self._loop
            assert self._cancel_event
            self._loop.call_soon_threadsafe(self._cancel_event.set)


class TaskGroup:
    """Object that `open_task_group()` yields."""

    def __init__(self) -> None:
        self._tasks: list[asyncio.Task] = []

    def start_soon(self, coro: Coroutine[Any, Any, Any]) -> None:
        """Schedule a task in the group.

        Args:
            coro: The return value of the `async` function defining the task.
        """
        self._tasks.append(asyncio.create_task(coro))

    async def _wait_all(self) -> None:
        """Block until all tasks complete.

        Raises:
            Exception: If one or more tasks raises an exception, one of these
                is raised arbitrarily.
        """
        done, _ = await asyncio.wait(
            self._tasks,
            # NOTE: Cancelling a task counts as a normal exit,
            #   not an exception.
            return_when=concurrent.futures.FIRST_EXCEPTION,
        )

        for task in done:
            try:
                if exc := task.exception():
                    raise exc
            except asyncio.CancelledError:
                pass

    def _cancel_all(self) -> None:
        """Cancel all tasks."""
        for task in self._tasks:
            # NOTE: It is safe to cancel tasks that have already completed.
            task.cancel()


@contextlib.asynccontextmanager
async def open_task_group() -> AsyncIterator[TaskGroup]:
    """Create a task group.

    `asyncio` gained task groups in Python 3.11.

    This is an async context manager, meant to be used with `async with`.
    On exit, it blocks until all subtasks complete. If any subtask fails, or if
    the current task is cancelled, it cancels all subtasks in the group and
    raises the subtask's exception. If multiple subtasks fail simultaneously,
    one of their exceptions is chosen arbitrarily.

    NOTE: Subtask exceptions do not propagate until the context manager exits.
    This means that the task group cannot cancel code running inside the
    `async with` block .
    """
    task_group = TaskGroup()

    try:
        yield task_group
        await task_group._wait_all()
    finally:
        task_group._cancel_all()


@contextlib.contextmanager
def cancel_on_exit(coro: Coroutine[Any, Any, Any]) -> Iterator[None]:
    """Schedule a task, cancelling it when exiting the context manager.

    If the given coroutine raises an exception, that exception is raised
    when exiting the context manager.
    """
    task = asyncio.create_task(coro)

    try:
        yield
    finally:
        if task.done() and (exception := task.exception()):
            raise exception

        task.cancel()



"""Storage handler."""

from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from wandb.sdk.lib.paths import FilePathStr, URIStr

if TYPE_CHECKING:
    from urllib.parse import ParseResult

    from wandb.sdk.artifacts.artifact import Artifact
    from wandb.sdk.artifacts.artifact_manifest_entry import ArtifactManifestEntry

DEFAULT_MAX_OBJECTS = 10**7


class StorageHandler:
    def can_handle(self, parsed_url: ParseResult) -> bool:
        """Checks whether this handler can handle the given url.

        Returns:
            Whether this handler can handle the given url.
        """
        raise NotImplementedError

    def load_path(
        self,
        manifest_entry: ArtifactManifestEntry,
        local: bool = False,
    ) -> URIStr | FilePathStr:
        """Load a file or directory given the corresponding index entry.

        Args:
            manifest_entry: The index entry to load
            local: Whether to load the file locally or not

        Returns:
            A path to the file represented by `index_entry`
        """
        raise NotImplementedError

    def store_path(
        self,
        artifact: Artifact,
        path: URIStr | FilePathStr,
        name: str | None = None,
        checksum: bool = True,
        max_objects: int | None = None,
    ) -> Sequence[ArtifactManifestEntry]:
        """Store the file or directory at the given path to the specified artifact.

        Args:
            path: The path to store
            name: If specified, the logical name that should map to `path`
            checksum: Whether to compute the checksum of the file
            max_objects: The maximum number of objects to store

        Returns:
            A list of manifest entries to store within the artifact
        """
        raise NotImplementedError



"""Utility to send an interrupt (Ctrl+C) signal to the main thread.

This is necessary because Windows and POSIX use different models for Ctrl+C
interrupts.
"""

import platform
import signal
import threading


def interrupt_main():
    """Interrupt the main Python thread with a SIGINT signal.

    In POSIX, signal.pthread_kill() is the most reliable way to send a signal
    to the main thread.

    os.kill() is often recommended, but it isn't guaranteed to deliver the
    signal to the main OS thread. Likewise, signal.raise_signal() delivers
    the signal to the current thread in POSIX. The issue is that if any other
    thread receives the signal, Python will set an internal flag and process it
    on the main thread at the next opportunity. If the main thread is executing
    C code or is blocked on a syscall (e.g. time.sleep(999999)) the signal
    handler won't execute until that's done---i.e. Python won't preempt the OS
    thread on its own.

    On Windows, pthread_kill is not available and os.kill() ignores its
    second argument and always kills the process. However,
    signal.raise_signal() does the right thing.
    """
    if platform.system() == "Windows":
        signal.raise_signal(signal.SIGINT)
    else:
        signal.pthread_kill(
            threading.main_thread().ident,
            signal.SIGINT,
        )



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

__all__ = [
    "ADD_ALIASES_GQL",
    "ARTIFACT_COLLECTION_MEMBERSHIP_FILES_GQL",
    "ARTIFACT_VERSION_FILES_GQL",
    "CREATE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL",
    "DELETE_ALIASES_GQL",
    "DELETE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL",
    "DELETE_ARTIFACT_PORTFOLIO_GQL",
    "DELETE_ARTIFACT_SEQUENCE_GQL",
    "FETCH_LINKED_ARTIFACTS_GQL",
    "MOVE_ARTIFACT_COLLECTION_GQL",
    "PROJECT_ARTIFACTS_GQL",
    "PROJECT_ARTIFACT_COLLECTIONS_GQL",
    "PROJECT_ARTIFACT_COLLECTION_GQL",
    "PROJECT_ARTIFACT_TYPES_GQL",
    "PROJECT_ARTIFACT_TYPE_GQL",
    "RUN_INPUT_ARTIFACTS_GQL",
    "RUN_OUTPUT_ARTIFACTS_GQL",
    "UPDATE_ARTIFACT_GQL",
    "UPDATE_ARTIFACT_PORTFOLIO_GQL",
    "UPDATE_ARTIFACT_SEQUENCE_GQL",
]

DELETE_ARTIFACT_SEQUENCE_GQL = """
mutation DeleteArtifactSequence($id: ID!) {
  deleteArtifactSequence(input: {artifactSequenceID: $id}) {
    artifactCollection {
      __typename
      state
    }
  }
}
"""

DELETE_ARTIFACT_PORTFOLIO_GQL = """
mutation DeleteArtifactPortfolio($id: ID!) {
  deleteArtifactPortfolio(input: {artifactPortfolioID: $id}) {
    artifactCollection {
      __typename
      state
    }
  }
}
"""

UPDATE_ARTIFACT_SEQUENCE_GQL = """
mutation UpdateArtifactSequence($id: ID!, $name: String, $description: String) {
  updateArtifactSequence(
    input: {artifactSequenceID: $id, name: $name, description: $description}
  ) {
    artifactCollection {
      __typename
      id
      name
      description
    }
  }
}
"""

UPDATE_ARTIFACT_PORTFOLIO_GQL = """
mutation UpdateArtifactPortfolio($id: ID!, $name: String, $description: String) {
  updateArtifactPortfolio(
    input: {artifactPortfolioID: $id, name: $name, description: $description}
  ) {
    artifactCollection {
      __typename
      id
      name
      description
    }
  }
}
"""

MOVE_ARTIFACT_COLLECTION_GQL = """
mutation MoveArtifactCollection($artifactSequenceID: ID!, $destinationArtifactTypeName: String!) {
  moveArtifactSequence(
    input: {artifactSequenceID: $artifactSequenceID, destinationArtifactTypeName: $destinationArtifactTypeName}
  ) {
    artifactCollection {
      __typename
      id
      name
      description
    }
  }
}
"""

CREATE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL = """
mutation CreateArtifactCollectionTagAssignments($entityName: String!, $projectName: String!, $artifactCollectionName: String!, $tags: [TagInput!]!) {
  createArtifactCollectionTagAssignments(
    input: {entityName: $entityName, projectName: $projectName, artifactCollectionName: $artifactCollectionName, tags: $tags}
  ) {
    tags {
      id
      name
      tagCategoryName
    }
  }
}
"""

DELETE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL = """
mutation DeleteArtifactCollectionTagAssignments($entityName: String!, $projectName: String!, $artifactCollectionName: String!, $tags: [TagInput!]!) {
  deleteArtifactCollectionTagAssignments(
    input: {entityName: $entityName, projectName: $projectName, artifactCollectionName: $artifactCollectionName, tags: $tags}
  ) {
    success
  }
}
"""

PROJECT_ARTIFACT_COLLECTIONS_GQL = """
query ProjectArtifactCollections($entityName: String!, $projectName: String!, $artifactTypeName: String!, $cursor: String) {
  project(name: $projectName, entityName: $entityName) {
    artifactType(name: $artifactTypeName) {
      artifactCollections: artifactCollections(after: $cursor) {
        ...ArtifactCollectionsFragment
      }
    }
  }
}

fragment ArtifactCollectionsFragment on ArtifactCollectionConnection {
  pageInfo {
    endCursor
    hasNextPage
  }
  totalCount
  edges {
    node {
      __typename
      id
      name
      description
      createdAt
    }
    cursor
  }
}
"""

PROJECT_ARTIFACT_COLLECTION_GQL = """
query ProjectArtifactCollection($entityName: String!, $projectName: String!, $artifactTypeName: String!, $artifactCollectionName: String!, $cursor: String, $perPage: Int = 1000) {
  project(name: $projectName, entityName: $entityName) {
    artifactType(name: $artifactTypeName) {
      artifactCollection: artifactCollection(name: $artifactCollectionName) {
        __typename
        id
        name
        description
        createdAt
        tags {
          edges {
            node {
              id
              name
            }
          }
        }
        aliases(after: $cursor, first: $perPage) {
          edges {
            node {
              alias
            }
            cursor
          }
          pageInfo {
            endCursor
            hasNextPage
          }
        }
      }
      artifactSequence(name: $artifactCollectionName) {
        __typename
      }
    }
  }
}
"""

ARTIFACT_VERSION_FILES_GQL = """
query ArtifactVersionFiles($entityName: String!, $projectName: String!, $artifactTypeName: String!, $artifactName: String!, $fileNames: [String!], $fileCursor: String, $fileLimit: Int = 50) {
  project(name: $projectName, entityName: $entityName) {
    artifactType(name: $artifactTypeName) {
      artifact(name: $artifactName) {
        files(names: $fileNames, after: $fileCursor, first: $fileLimit) {
          ...FilesFragment
        }
      }
    }
  }
}

fragment FilesFragment on FileConnection {
  edges {
    node {
      id
      name: displayName
      url
      sizeBytes
      storagePath @include(if: true)
      mimetype
      updatedAt
      digest
      md5
      directUrl
    }
    cursor
  }
  pageInfo {
    endCursor
    hasNextPage
  }
}
"""

ARTIFACT_COLLECTION_MEMBERSHIP_FILES_GQL = """
query ArtifactCollectionMembershipFiles($entityName: String!, $projectName: String!, $artifactName: String!, $artifactVersionIndex: String!, $fileNames: [String!], $fileCursor: String, $fileLimit: Int = 50) {
  project(name: $projectName, entityName: $entityName) {
    artifactCollection(name: $artifactName) {
      __typename
      artifactMembership(aliasName: $artifactVersionIndex) {
        files(names: $fileNames, after: $fileCursor, first: $fileLimit) {
          ...FilesFragment
        }
      }
    }
  }
}

fragment FilesFragment on FileConnection {
  edges {
    node {
      id
      name: displayName
      url
      sizeBytes
      storagePath @include(if: true)
      mimetype
      updatedAt
      digest
      md5
      directUrl
    }
    cursor
  }
  pageInfo {
    endCursor
    hasNextPage
  }
}
"""

PROJECT_ARTIFACT_TYPES_GQL = """
query ProjectArtifactTypes($entityName: String!, $projectName: String!, $cursor: String) {
  project(name: $projectName, entityName: $entityName) {
    artifactTypes(after: $cursor) {
      ...ArtifactTypesFragment
    }
  }
}

fragment ArtifactTypeFragment on ArtifactType {
  __typename
  id
  name
  description
  createdAt
}

fragment ArtifactTypesFragment on ArtifactTypeConnection {
  edges {
    node {
      ...ArtifactTypeFragment
    }
    cursor
  }
  pageInfo {
    endCursor
    hasNextPage
  }
}
"""

PROJECT_ARTIFACT_TYPE_GQL = """
query ProjectArtifactType($entityName: String!, $projectName: String!, $artifactTypeName: String!) {
  project(name: $projectName, entityName: $entityName) {
    artifactType(name: $artifactTypeName) {
      ...ArtifactTypeFragment
    }
  }
}

fragment ArtifactTypeFragment on ArtifactType {
  __typename
  id
  name
  description
  createdAt
}
"""

PROJECT_ARTIFACTS_GQL = """
query ProjectArtifacts($project: String!, $entity: String!, $type: String!, $collection: String!, $cursor: String, $perPage: Int = 50, $order: String, $filters: JSONString) {
  project(name: $project, entityName: $entity) {
    artifactType(name: $type) {
      artifactCollection: artifactCollection(name: $collection) {
        __typename
        name
        artifacts(filters: $filters, after: $cursor, first: $perPage, order: $order) {
          ...ArtifactsFragment
        }
      }
    }
  }
}

fragment ArtifactFragment on Artifact {
  id
  artifactSequence {
    project {
      entityName
      name
    }
    name
  }
  versionIndex
  artifactType {
    name
  }
  description
  metadata
  ttlDurationSeconds @include(if: true)
  ttlIsInherited @include(if: true)
  aliases @include(if: true) {
    artifactCollection {
      __typename
      project {
        entityName
        name
      }
      name
    }
    alias
  }
  tags @include(if: true) {
    name
  }
  historyStep @include(if: true)
  state
  currentManifest {
    file {
      directUrl
    }
  }
  commitHash
  fileCount
  createdAt
  updatedAt
}

fragment ArtifactsFragment on VersionedArtifactConnection {
  totalCount
  edges {
    node {
      ...ArtifactFragment
    }
    version
    cursor
  }
  pageInfo {
    endCursor
    hasNextPage
  }
}
"""

RUN_OUTPUT_ARTIFACTS_GQL = """
query RunOutputArtifacts($entity: String!, $project: String!, $runName: String!, $cursor: String, $perPage: Int) {
  project(name: $project, entityName: $entity) {
    run(name: $runName) {
      outputArtifacts(after: $cursor, first: $perPage) {
        totalCount
        edges {
          node {
            ...ArtifactFragment
          }
          cursor
        }
        pageInfo {
          endCursor
          hasNextPage
        }
      }
    }
  }
}

fragment ArtifactFragment on Artifact {
  id
  artifactSequence {
    project {
      entityName
      name
    }
    name
  }
  versionIndex
  artifactType {
    name
  }
  description
  metadata
  ttlDurationSeconds @include(if: true)
  ttlIsInherited @include(if: true)
  aliases @include(if: true) {
    artifactCollection {
      __typename
      project {
        entityName
        name
      }
      name
    }
    alias
  }
  tags @include(if: true) {
    name
  }
  historyStep @include(if: true)
  state
  currentManifest {
    file {
      directUrl
    }
  }
  commitHash
  fileCount
  createdAt
  updatedAt
}
"""

RUN_INPUT_ARTIFACTS_GQL = """
query RunInputArtifacts($entity: String!, $project: String!, $runName: String!, $cursor: String, $perPage: Int) {
  project(name: $project, entityName: $entity) {
    run(name: $runName) {
      inputArtifacts(after: $cursor, first: $perPage) {
        totalCount
        edges {
          node {
            ...ArtifactFragment
          }
          cursor
        }
        pageInfo {
          endCursor
          hasNextPage
        }
      }
    }
  }
}

fragment ArtifactFragment on Artifact {
  id
  artifactSequence {
    project {
      entityName
      name
    }
    name
  }
  versionIndex
  artifactType {
    name
  }
  description
  metadata
  ttlDurationSeconds @include(if: true)
  ttlIsInherited @include(if: true)
  aliases @include(if: true) {
    artifactCollection {
      __typename
      project {
        entityName
        name
      }
      name
    }
    alias
  }
  tags @include(if: true) {
    name
  }
  historyStep @include(if: true)
  state
  currentManifest {
    file {
      directUrl
    }
  }
  commitHash
  fileCount
  createdAt
  updatedAt
}
"""

FETCH_LINKED_ARTIFACTS_GQL = """
query FetchLinkedArtifacts($artifactID: ID!) {
  artifact(id: $artifactID) {
    artifactMemberships {
      edges {
        node {
          aliases {
            alias
          }
          versionIndex
          artifactCollection {
            project {
              entityName
              name
            }
            name
            __typename
          }
        }
      }
    }
  }
}
"""

ADD_ALIASES_GQL = """
mutation AddAliases($artifactID: ID!, $aliases: [ArtifactCollectionAliasInput!]!) {
  addAliases(input: {artifactID: $artifactID, aliases: $aliases}) {
    success
  }
}
"""

DELETE_ALIASES_GQL = """
mutation DeleteAliases($artifactID: ID!, $aliases: [ArtifactCollectionAliasInput!]!) {
  deleteAliases(input: {artifactID: $artifactID, aliases: $aliases}) {
    success
  }
}
"""

UPDATE_ARTIFACT_GQL = """
mutation UpdateArtifact($artifactID: ID!, $description: String, $metadata: JSONString, $ttlDurationSeconds: Int64, $tagsToAdd: [TagInput!], $tagsToDelete: [TagInput!], $aliases: [ArtifactAliasInput!]) {
  updateArtifact(
    input: {artifactID: $artifactID, description: $description, metadata: $metadata, ttlDurationSeconds: $ttlDurationSeconds, tagsToAdd: $tagsToAdd, tagsToDelete: $tagsToDelete, aliases: $aliases}
  ) {
    artifact {
      ...ArtifactFragment
    }
  }
}

fragment ArtifactFragment on Artifact {
  id
  artifactSequence {
    project {
      entityName
      name
    }
    name
  }
  versionIndex
  artifactType {
    name
  }
  description
  metadata
  ttlDurationSeconds @include(if: true)
  ttlIsInherited @include(if: true)
  aliases @include(if: true) {
    artifactCollection {
      __typename
      project {
        entityName
        name
      }
      name
    }
    alias
  }
  tags @include(if: true) {
    name
  }
  historyStep @include(if: true)
  state
  currentManifest {
    file {
      directUrl
    }
  }
  commitHash
  fileCount
  createdAt
  updatedAt
}
"""



"""wandb integration sagemaker module."""

from .auth import sagemaker_auth
from .config import is_using_sagemaker, parse_sm_config
from .resources import parse_sm_secrets, set_global_settings, set_run_id

__all__ = [
    "sagemaker_auth",
    "is_using_sagemaker",
    "parse_sm_config",
    "parse_sm_secrets",
    "set_global_settings",
    "set_run_id",
]



import wandb
from wandb._pydantic import IS_PYDANTIC_V2

from .actions import ActionType, DoNothing, SendNotification, SendWebhook
from .automations import Automation, NewAutomation
from .events import (
    ArtifactEvent,
    EventType,
    MetricChangeFilter,
    MetricThresholdFilter,
    OnAddArtifactAlias,
    OnCreateArtifact,
    OnLinkArtifact,
    OnRunMetric,
    RunEvent,
)
from .integrations import Integration, SlackIntegration, WebhookIntegration
from .scopes import ArtifactCollectionScope, ProjectScope, ScopeType

# ----------------------------------------------------------------------------
# WARNINGS on import
if not IS_PYDANTIC_V2:
    # Raises an error in Pydantic v1 environments, where the Automations API
    # has not been tested and is unlikely to work as expected.
    #
    # Remove this when we either:
    # - Drop support for Pydantic v1
    # - Are able to implement (limited) Pydantic v1 support
    raise ImportError(
        "The W&B Automations API requires Pydantic v2. "
        "We recommend upgrading `pydantic` to use this feature."
    )

else:
    # If Pydantic v2 is available, we can use the full Automations API
    # but communicate to users that the API is still experimental and
    # may change rapidly.
    wandb.termwarn(
        "The W&B Automations API is experimental and the implementation is subject to change."
        "Review the release notes before upgrading. We recommend pinning your "
        f"package version to `{wandb.__package__}=={wandb.__version__}` to reduce the risk of disruption.",
        repeat=False,
    )
# ----------------------------------------------------------------------------

__all__ = [
    # Scopes
    "ScopeType",
    "ArtifactCollectionScope",
    "ProjectScope",
    # Events
    "EventType",
    "OnAddArtifactAlias",
    "OnCreateArtifact",
    "OnLinkArtifact",
    "OnRunMetric",
    "ArtifactEvent",
    "RunEvent",
    "MetricThresholdFilter",
    "MetricChangeFilter",
    # Actions
    "ActionType",
    "SendNotification",
    "SendWebhook",
    "DoNothing",
    # Automations
    "Automation",
    "NewAutomation",
    # Integrations
    "Integration",
    "SlackIntegration",
    "WebhookIntegration",
]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Literal, Optional

from pydantic import Field

from wandb._pydantic import GQLBase, GQLId, Typename


class MoveArtifactCollection(GQLBase):
    move_artifact_sequence: Optional[MoveArtifactCollectionMoveArtifactSequence] = (
        Field(alias="moveArtifactSequence")
    )


class MoveArtifactCollectionMoveArtifactSequence(GQLBase):
    artifact_collection: Optional[
        MoveArtifactCollectionMoveArtifactSequenceArtifactCollection
    ] = Field(alias="artifactCollection")


class MoveArtifactCollectionMoveArtifactSequenceArtifactCollection(GQLBase):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    id: GQLId
    name: str
    description: Optional[str]


MoveArtifactCollection.model_rebuild()
MoveArtifactCollectionMoveArtifactSequence.model_rebuild()



"""watch."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Sequence

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

import wandb

from .lib import telemetry

if TYPE_CHECKING:
    import torch  # type: ignore [import-not-found]

    from wandb.sdk.wandb_run import Run

logger = logging.getLogger("wandb")

_global_watch_idx = 0


def _watch(
    run: Run,
    models: torch.nn.Module | Sequence[torch.nn.Module],
    criterion: torch.F | None = None,
    log: Literal["gradients", "parameters", "all"] | None = "gradients",
    log_freq: int = 1000,
    idx: int | None = None,
    log_graph: bool = False,
):
    """Hooks into the given PyTorch model(s) to monitor gradients and the model's computational graph.

    This function can track parameters, gradients, or both during training. It should be
    extended to support arbitrary machine learning models in the future.

    Args:
        run (wandb.sdk.wandb_run.Run): The run object to log to.
        models (Union[torch.nn.Module, Sequence[torch.nn.Module]]):
            A single model or a sequence of models to be monitored.
        criterion (Optional[torch.F]):
            The loss function being optimized (optional).
        log (Optional[Literal["gradients", "parameters", "all"]]):
            Specifies whether to log "gradients", "parameters", or "all".
            Set to None to disable logging. (default="gradients")
        log_freq (int):
            Frequency (in batches) to log gradients and parameters. (default=1000)
        idx (Optional[int]):
            Index used when tracking multiple models with `wandb.watch`. (default=None)
         log_graph (bool):
            Whether to log the model's computational graph. (default=False)

    Returns:
        wandb.Graph:
            The graph object, which will be populated after the first backward pass.

    Raises:
        ValueError:
            If `wandb.init` has not been called or if any of the models are not instances
            of `torch.nn.Module`.
    """
    global _global_watch_idx

    with telemetry.context() as tel:
        tel.feature.watch = True

    logger.info("Watching")

    if log not in {"gradients", "parameters", "all", None}:
        raise ValueError("log must be one of 'gradients', 'parameters', 'all', or None")

    log_parameters = log in {"parameters", "all"}
    log_gradients = log in {"gradients", "all"}

    if not isinstance(models, (tuple, list)):
        models = (models,)

    torch = wandb.util.get_module(
        "torch", required="wandb.watch only works with pytorch, couldn't import torch."
    )

    for model in models:
        if not isinstance(model, torch.nn.Module):
            raise ValueError(
                "Expected a pytorch model (torch.nn.Module). Received "
                + str(type(model))
            )

    graphs = []
    prefix = ""

    if idx is None:
        idx = _global_watch_idx
    for local_idx, model in enumerate(models):
        global_idx = idx + local_idx
        _global_watch_idx += 1
        if global_idx > 0:
            # TODO: this makes ugly chart names like gradients/graph_1conv1d.bias
            prefix = f"graph_{global_idx}"

        if log_parameters:
            run._torch.add_log_parameters_hook(
                model,
                prefix=prefix,
                log_freq=log_freq,
            )

        if log_gradients:
            run._torch.add_log_gradients_hook(
                model,
                prefix=prefix,
                log_freq=log_freq,
            )

        if log_graph:
            graph = run._torch.hook_torch(model, criterion, graph_idx=global_idx)
            graphs.append(graph)
            # NOTE: the graph is set in run.summary by hook_torch on the backward pass
    return graphs


def _unwatch(
    run: Run, models: torch.nn.Module | Sequence[torch.nn.Module] | None = None
) -> None:
    """Remove pytorch model topology, gradient and parameter hooks.

    Args:
        run (wandb.sdk.wandb_run.Run):
            The run object to log to.
        models (torch.nn.Module | Sequence[torch.nn.Module]):
            Optional list of pytorch models that have had watch called on them
    """
    if models:
        if not isinstance(models, (tuple, list)):
            models = (models,)
        for model in models:
            if not hasattr(model, "_wandb_hook_names"):
                wandb.termwarn("{} model has not been watched".format(model))
            else:
                for name in model._wandb_hook_names:
                    run._torch.unhook(name)
                delattr(model, "_wandb_hook_names")
                # TODO: we should also remove recursively model._wandb_watch_called

    else:
        run._torch.unhook_all()



"""Router - handle message router (relay).

Router to manage responses from a queue with relay.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_server_pb2 as spb
from wandb.sdk.mailbox import Mailbox

from .router_queue import MessageQueueRouter

if TYPE_CHECKING:
    from queue import Queue


class MessageRelayRouter(MessageQueueRouter):
    _relay_queue: Queue[pb.Result]

    def __init__(
        self,
        request_queue: Queue[pb.Record],
        response_queue: Queue[pb.Result],
        relay_queue: Queue[pb.Result],
        mailbox: Mailbox,
    ) -> None:
        self._relay_queue = relay_queue
        super().__init__(
            request_queue=request_queue,
            response_queue=response_queue,
            mailbox=mailbox,
        )

    def _handle_msg_rcv(self, msg: pb.Result | spb.ServerResponse) -> None:
        if isinstance(msg, pb.Result):
            relay_msg = msg
        else:
            relay_msg = msg.result_communicate

        # This is legacy-service logic for returning responses to the client.
        # A different thread reads the "relay queue" and writes responses on
        # the socket.
        if relay_msg.control.relay_id:
            self._relay_queue.put(relay_msg)
        else:
            super()._handle_msg_rcv(msg)



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from datetime import datetime
from typing import List, Literal, Optional, Union

from pydantic import Field
from typing_extensions import Annotated

from wandb._pydantic import GQLBase, GQLId, Typename

from .enums import AlertSeverity, EventTriggeringConditionType


class ArtifactPortfolioScopeFields(GQLBase):
    typename__: Typename[Literal["ArtifactPortfolio"]] = "ArtifactPortfolio"
    id: GQLId
    name: str


class ArtifactSequenceScopeFields(GQLBase):
    typename__: Typename[Literal["ArtifactSequence"]] = "ArtifactSequence"
    id: GQLId
    name: str


class CreateAutomationResult(GQLBase):
    typename__: Typename[Literal["CreateFilterTriggerPayload"]] = (
        "CreateFilterTriggerPayload"
    )
    trigger: Optional[TriggerFields]


class DeleteAutomationResult(GQLBase):
    typename__: Typename[Literal["DeleteTriggerPayload"]] = "DeleteTriggerPayload"
    success: bool


class FilterEventFields(GQLBase):
    typename__: Typename[Literal["FilterEventTriggeringCondition"]] = (
        "FilterEventTriggeringCondition"
    )
    event_type: EventTriggeringConditionType = Field(alias="eventType")
    filter: str


class GenericWebhookActionFields(GQLBase):
    typename__: Typename[Literal["GenericWebhookTriggeredAction"]] = (
        "GenericWebhookTriggeredAction"
    )
    integration: Union[
        GenericWebhookActionFieldsIntegrationIntegration,
        GenericWebhookActionFieldsIntegrationGenericWebhookIntegration,
    ] = Field(discriminator="typename__")
    request_payload: Optional[str] = Field(alias="requestPayload")


class GenericWebhookActionFieldsIntegrationIntegration(GQLBase):
    typename__: Typename[
        Literal["GitHubOAuthIntegration", "Integration", "SlackIntegration"]
    ]


class GenericWebhookIntegrationConnectionFields(GQLBase):
    typename__: Typename[Literal["IntegrationConnection"]] = "IntegrationConnection"
    page_info: PageInfoFields = Field(alias="pageInfo")
    edges: List[GenericWebhookIntegrationConnectionFieldsEdges]


class GenericWebhookIntegrationConnectionFieldsEdges(GQLBase):
    cursor: str
    node: Optional[
        Annotated[
            Union[
                GenericWebhookIntegrationConnectionFieldsEdgesNodeIntegration,
                GenericWebhookIntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration,
            ],
            Field(discriminator="typename__"),
        ]
    ]


class GenericWebhookIntegrationConnectionFieldsEdgesNodeIntegration(GQLBase):
    typename__: Typename[
        Literal["GitHubOAuthIntegration", "Integration", "SlackIntegration"]
    ]


class GenericWebhookIntegrationFields(GQLBase):
    typename__: Typename[Literal["GenericWebhookIntegration"]] = (
        "GenericWebhookIntegration"
    )
    id: GQLId
    name: str
    url_endpoint: str = Field(alias="urlEndpoint")


class IntegrationConnectionFields(GQLBase):
    typename__: Typename[Literal["IntegrationConnection"]] = "IntegrationConnection"
    page_info: PageInfoFields = Field(alias="pageInfo")
    edges: List[IntegrationConnectionFieldsEdges]


class IntegrationConnectionFieldsEdges(GQLBase):
    cursor: str
    node: Optional[
        Annotated[
            Union[
                IntegrationConnectionFieldsEdgesNodeIntegration,
                IntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration,
                IntegrationConnectionFieldsEdgesNodeSlackIntegration,
            ],
            Field(discriminator="typename__"),
        ]
    ]


class IntegrationConnectionFieldsEdgesNodeIntegration(GQLBase):
    typename__: Typename[Literal["GitHubOAuthIntegration", "Integration"]]


class NoOpActionFields(GQLBase):
    typename__: Typename[Literal["NoOpTriggeredAction"]] = "NoOpTriggeredAction"
    no_op: Optional[bool] = Field(alias="noOp")


class NotificationActionFields(GQLBase):
    typename__: Typename[Literal["NotificationTriggeredAction"]] = (
        "NotificationTriggeredAction"
    )
    integration: Union[
        NotificationActionFieldsIntegrationIntegration,
        NotificationActionFieldsIntegrationSlackIntegration,
    ] = Field(discriminator="typename__")
    title: Optional[str]
    message: Optional[str]
    severity: Optional[AlertSeverity]


class NotificationActionFieldsIntegrationIntegration(GQLBase):
    typename__: Typename[
        Literal["GenericWebhookIntegration", "GitHubOAuthIntegration", "Integration"]
    ]


class PageInfoFields(GQLBase):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


class ProjectConnectionFields(GQLBase):
    typename__: Typename[Literal["ProjectConnection"]] = "ProjectConnection"
    page_info: PageInfoFields = Field(alias="pageInfo")
    edges: List[ProjectConnectionFieldsEdges]


class ProjectConnectionFieldsEdges(GQLBase):
    cursor: str
    node: Optional[ProjectConnectionFieldsEdgesNode]


class ProjectConnectionFieldsEdgesNode(GQLBase):
    triggers: List[TriggerFields]


class ProjectScopeFields(GQLBase):
    typename__: Typename[Literal["Project"]] = "Project"
    id: GQLId
    name: str


class QueueJobActionFields(GQLBase):
    typename__: Typename[Literal["QueueJobTriggeredAction"]] = "QueueJobTriggeredAction"
    queue: Optional[QueueJobActionFieldsQueue]
    template: str


class QueueJobActionFieldsQueue(GQLBase):
    id: GQLId
    name: str


class SlackIntegrationConnectionFields(GQLBase):
    typename__: Typename[Literal["IntegrationConnection"]] = "IntegrationConnection"
    page_info: PageInfoFields = Field(alias="pageInfo")
    edges: List[SlackIntegrationConnectionFieldsEdges]


class SlackIntegrationConnectionFieldsEdges(GQLBase):
    cursor: str
    node: Optional[
        Annotated[
            Union[
                SlackIntegrationConnectionFieldsEdgesNodeIntegration,
                SlackIntegrationConnectionFieldsEdgesNodeSlackIntegration,
            ],
            Field(discriminator="typename__"),
        ]
    ]


class SlackIntegrationConnectionFieldsEdgesNodeIntegration(GQLBase):
    typename__: Typename[
        Literal["GenericWebhookIntegration", "GitHubOAuthIntegration", "Integration"]
    ]


class SlackIntegrationFields(GQLBase):
    typename__: Typename[Literal["SlackIntegration"]] = "SlackIntegration"
    id: GQLId
    team_name: str = Field(alias="teamName")
    channel_name: str = Field(alias="channelName")


class TriggerFields(GQLBase):
    typename__: Typename[Literal["Trigger"]] = "Trigger"
    id: GQLId
    created_at: datetime = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(alias="updatedAt")
    name: str
    description: Optional[str]
    enabled: bool
    scope: Union[
        TriggerFieldsScopeProject,
        TriggerFieldsScopeArtifactSequence,
        TriggerFieldsScopeArtifactPortfolio,
    ] = Field(discriminator="typename__")
    event: TriggerFieldsEventFilterEventTriggeringCondition
    action: Union[
        TriggerFieldsActionQueueJobTriggeredAction,
        TriggerFieldsActionNotificationTriggeredAction,
        TriggerFieldsActionGenericWebhookTriggeredAction,
        TriggerFieldsActionNoOpTriggeredAction,
    ] = Field(discriminator="typename__")


class UpdateAutomationResult(GQLBase):
    typename__: Typename[Literal["UpdateFilterTriggerPayload"]] = (
        "UpdateFilterTriggerPayload"
    )
    trigger: Optional[TriggerFields]


class TriggerFieldsScopeArtifactPortfolio(ArtifactPortfolioScopeFields):
    typename__: Typename[Literal["ArtifactPortfolio"]]


class TriggerFieldsScopeArtifactSequence(ArtifactSequenceScopeFields):
    typename__: Typename[Literal["ArtifactSequence"]]


class TriggerFieldsEventFilterEventTriggeringCondition(FilterEventFields):
    typename__: Typename[Literal["FilterEventTriggeringCondition"]]


class TriggerFieldsActionGenericWebhookTriggeredAction(GenericWebhookActionFields):
    typename__: Typename[Literal["GenericWebhookTriggeredAction"]]


class GenericWebhookActionFieldsIntegrationGenericWebhookIntegration(
    GenericWebhookIntegrationFields
):
    typename__: Typename[Literal["GenericWebhookIntegration"]]


class GenericWebhookIntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration(
    GenericWebhookIntegrationFields
):
    typename__: Typename[Literal["GenericWebhookIntegration"]]


class IntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration(
    GenericWebhookIntegrationFields
):
    typename__: Typename[Literal["GenericWebhookIntegration"]]


class TriggerFieldsActionNoOpTriggeredAction(NoOpActionFields):
    typename__: Typename[Literal["NoOpTriggeredAction"]]


class TriggerFieldsActionNotificationTriggeredAction(NotificationActionFields):
    typename__: Typename[Literal["NotificationTriggeredAction"]]


class TriggerFieldsScopeProject(ProjectScopeFields):
    typename__: Typename[Literal["Project"]]


class TriggerFieldsActionQueueJobTriggeredAction(QueueJobActionFields):
    typename__: Typename[Literal["QueueJobTriggeredAction"]]


class IntegrationConnectionFieldsEdgesNodeSlackIntegration(SlackIntegrationFields):
    typename__: Typename[Literal["SlackIntegration"]]


class NotificationActionFieldsIntegrationSlackIntegration(SlackIntegrationFields):
    typename__: Typename[Literal["SlackIntegration"]]


class SlackIntegrationConnectionFieldsEdgesNodeSlackIntegration(SlackIntegrationFields):
    typename__: Typename[Literal["SlackIntegration"]]


ArtifactPortfolioScopeFields.model_rebuild()
ArtifactSequenceScopeFields.model_rebuild()
CreateAutomationResult.model_rebuild()
DeleteAutomationResult.model_rebuild()
FilterEventFields.model_rebuild()
GenericWebhookActionFields.model_rebuild()
GenericWebhookActionFieldsIntegrationIntegration.model_rebuild()
GenericWebhookIntegrationConnectionFields.model_rebuild()
GenericWebhookIntegrationConnectionFieldsEdges.model_rebuild()
GenericWebhookIntegrationConnectionFieldsEdgesNodeIntegration.model_rebuild()
GenericWebhookIntegrationFields.model_rebuild()
IntegrationConnectionFields.model_rebuild()
IntegrationConnectionFieldsEdges.model_rebuild()
IntegrationConnectionFieldsEdgesNodeIntegration.model_rebuild()
NoOpActionFields.model_rebuild()
NotificationActionFields.model_rebuild()
NotificationActionFieldsIntegrationIntegration.model_rebuild()
PageInfoFields.model_rebuild()
ProjectConnectionFields.model_rebuild()
ProjectConnectionFieldsEdges.model_rebuild()
ProjectConnectionFieldsEdgesNode.model_rebuild()
ProjectScopeFields.model_rebuild()
QueueJobActionFields.model_rebuild()
QueueJobActionFieldsQueue.model_rebuild()
SlackIntegrationConnectionFields.model_rebuild()
SlackIntegrationConnectionFieldsEdges.model_rebuild()
SlackIntegrationConnectionFieldsEdgesNodeIntegration.model_rebuild()
SlackIntegrationFields.model_rebuild()
TriggerFields.model_rebuild()
UpdateAutomationResult.model_rebuild()
TriggerFieldsScopeArtifactPortfolio.model_rebuild()
TriggerFieldsScopeArtifactSequence.model_rebuild()
TriggerFieldsEventFilterEventTriggeringCondition.model_rebuild()
TriggerFieldsActionGenericWebhookTriggeredAction.model_rebuild()
GenericWebhookActionFieldsIntegrationGenericWebhookIntegration.model_rebuild()
GenericWebhookIntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration.model_rebuild()
IntegrationConnectionFieldsEdgesNodeGenericWebhookIntegration.model_rebuild()
TriggerFieldsActionNoOpTriggeredAction.model_rebuild()
TriggerFieldsActionNotificationTriggeredAction.model_rebuild()
PageInfoFields.model_rebuild()
PageInfoFields.model_rebuild()
PageInfoFields.model_rebuild()
PageInfoFields.model_rebuild()
TriggerFieldsScopeProject.model_rebuild()
TriggerFieldsActionQueueJobTriggeredAction.model_rebuild()
IntegrationConnectionFieldsEdgesNodeSlackIntegration.model_rebuild()
NotificationActionFieldsIntegrationSlackIntegration.model_rebuild()
SlackIntegrationConnectionFieldsEdgesNodeSlackIntegration.model_rebuild()
TriggerFields.model_rebuild()
TriggerFields.model_rebuild()
TriggerFields.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import ArtifactTypesFragment


class ProjectArtifactTypes(GQLBase):
    project: Optional[ProjectArtifactTypesProject]


class ProjectArtifactTypesProject(GQLBase):
    artifact_types: ArtifactTypesFragment = Field(alias="artifactTypes")


ProjectArtifactTypes.model_rebuild()
ProjectArtifactTypesProject.model_rebuild()



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any, Iterable, Mapping

from pydantic import ValidationError
from typing_extensions import override
from wandb_graphql.language.ast import Document

from wandb.apis.paginator import Paginator, _Client

if TYPE_CHECKING:
    from wandb.automations import Automation
    from wandb.automations._generated import ProjectConnectionFields


class Automations(Paginator["Automation"]):
    last_response: ProjectConnectionFields | None
    _query: Document

    def __init__(
        self,
        client: _Client,
        variables: Mapping[str, Any],
        per_page: int = 50,
        _query: Document | None = None,
    ):
        super().__init__(client, variables, per_page=per_page)
        if _query is None:
            raise RuntimeError(f"Query required for {type(self).__qualname__}")
        self._query = _query

    @property
    def more(self) -> bool:
        """Whether there are more items to fetch."""
        if self.last_response is None:
            return True
        return self.last_response.page_info.has_next_page

    @property
    def cursor(self) -> str | None:
        """The start cursor to use for the next page."""
        if self.last_response is None:
            return None
        return self.last_response.page_info.end_cursor

    @override
    def _update_response(self) -> None:
        """Fetch the raw response data for the current page."""
        from wandb.automations._generated import ProjectConnectionFields

        data: dict[str, Any] = self.client.execute(
            self._query, variable_values=self.variables
        )
        try:
            page_data = data["searchScope"]["projects"]
            self.last_response = ProjectConnectionFields.model_validate(page_data)
        except (LookupError, AttributeError, ValidationError) as e:
            raise ValueError("Unexpected response data") from e

    def convert_objects(self) -> Iterable[Automation]:
        """Parse the page data into a list of objects."""
        from wandb.automations import Automation

        page = self.last_response
        return [
            Automation.model_validate(obj)
            for obj in chain.from_iterable(edge.node.triggers for edge in page.edges)
        ]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import ProjectConnectionFields


class GetAutomations(GQLBase):
    search_scope: Optional[GetAutomationsSearchScope] = Field(alias="searchScope")


class GetAutomationsSearchScope(GQLBase):
    projects: Optional[ProjectConnectionFields]


GetAutomations.model_rebuild()
GetAutomationsSearchScope.model_rebuild()



# Generated by ariadne-codegen

from .delete_project import DeleteProject, DeleteProjectDeleteModel
from .fetch_registry import FetchRegistry, FetchRegistryEntity
from .fragments import (
    RegistryFragment,
    RegistryFragmentArtifactTypes,
    RegistryFragmentArtifactTypesEdges,
    RegistryFragmentArtifactTypesEdgesNode,
)
from .input_types import ArtifactTypeInput
from .operations import (
    DELETE_PROJECT_GQL,
    FETCH_REGISTRY_GQL,
    RENAME_PROJECT_GQL,
    UPSERT_REGISTRY_PROJECT_GQL,
)
from .rename_project import (
    RenameProject,
    RenameProjectRenameProject,
    RenameProjectRenameProjectProject,
)
from .upsert_registry_project import (
    UpsertRegistryProject,
    UpsertRegistryProjectUpsertModel,
)

__all__ = [
    "DELETE_PROJECT_GQL",
    "FETCH_REGISTRY_GQL",
    "RENAME_PROJECT_GQL",
    "UPSERT_REGISTRY_PROJECT_GQL",
    "FetchRegistry",
    "FetchRegistryEntity",
    "RenameProject",
    "RenameProjectRenameProject",
    "RenameProjectRenameProjectProject",
    "UpsertRegistryProject",
    "UpsertRegistryProjectUpsertModel",
    "DeleteProject",
    "DeleteProjectDeleteModel",
    "ArtifactTypeInput",
    "RegistryFragment",
    "RegistryFragmentArtifactTypes",
    "RegistryFragmentArtifactTypesEdges",
    "RegistryFragmentArtifactTypesEdgesNode",
]



"""InterfaceSock - Derived from InterfaceShared using a socket to send to internal thread.

See interface.py for how interface classes relate to each other.

"""

import logging
from typing import TYPE_CHECKING, Any, Optional

from wandb.sdk.mailbox import Mailbox

from ..lib.sock_client import SockClient
from .interface_shared import InterfaceShared

if TYPE_CHECKING:
    from wandb.proto import wandb_internal_pb2 as pb


logger = logging.getLogger("wandb")


class InterfaceSock(InterfaceShared):
    def __init__(
        self,
        sock_client: SockClient,
        mailbox: Mailbox,
        stream_id: str,
    ) -> None:
        super().__init__(mailbox=mailbox)
        self._sock_client = sock_client
        self._stream_id = stream_id

    def _assign(self, record: Any) -> None:
        assert self._stream_id
        record._info.stream_id = self._stream_id

    def _publish(self, record: "pb.Record", local: Optional[bool] = None) -> None:
        self._assign(record)
        self._sock_client.send_record_publish(record)



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Any, List, Literal, Optional

from pydantic import Field

from wandb._pydantic import GQLBase, GQLId, Typename

from .enums import ArtifactState


class ArtifactCollectionsFragment(GQLBase):
    page_info: ArtifactCollectionsFragmentPageInfo = Field(alias="pageInfo")
    total_count: int = Field(alias="totalCount")
    edges: List[ArtifactCollectionsFragmentEdges]


class ArtifactCollectionsFragmentEdges(GQLBase):
    node: Optional[ArtifactCollectionsFragmentEdgesNode]
    cursor: str


class ArtifactCollectionsFragmentEdgesNode(GQLBase):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    id: GQLId
    name: str
    description: Optional[str]
    created_at: str = Field(alias="createdAt")


class ArtifactCollectionsFragmentPageInfo(GQLBase):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


class ArtifactFragment(GQLBase):
    id: GQLId
    artifact_sequence: ArtifactFragmentArtifactSequence = Field(
        alias="artifactSequence"
    )
    version_index: Optional[int] = Field(alias="versionIndex")
    artifact_type: ArtifactFragmentArtifactType = Field(alias="artifactType")
    description: Optional[str]
    metadata: Optional[Any]
    ttl_duration_seconds: Optional[Any] = Field(
        alias="ttlDurationSeconds", default=None
    )
    ttl_is_inherited: Optional[bool] = Field(alias="ttlIsInherited", default=None)
    aliases: Optional[List[ArtifactFragmentAliases]] = None
    tags: Optional[List[ArtifactFragmentTags]] = None
    history_step: Optional[Any] = Field(alias="historyStep", default=None)
    state: ArtifactState
    current_manifest: Optional[ArtifactFragmentCurrentManifest] = Field(
        alias="currentManifest"
    )
    commit_hash: Optional[str] = Field(alias="commitHash")
    file_count: Any = Field(alias="fileCount")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")


class ArtifactFragmentAliases(GQLBase):
    artifact_collection: Optional[ArtifactFragmentAliasesArtifactCollection] = Field(
        alias="artifactCollection"
    )
    alias: str


class ArtifactFragmentAliasesArtifactCollection(GQLBase):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    project: Optional[ArtifactFragmentAliasesArtifactCollectionProject]
    name: str


class ArtifactFragmentAliasesArtifactCollectionProject(GQLBase):
    entity_name: str = Field(alias="entityName")
    name: str


class ArtifactFragmentArtifactSequence(GQLBase):
    project: Optional[ArtifactFragmentArtifactSequenceProject]
    name: str


class ArtifactFragmentArtifactSequenceProject(GQLBase):
    entity_name: str = Field(alias="entityName")
    name: str


class ArtifactFragmentArtifactType(GQLBase):
    name: str


class ArtifactFragmentCurrentManifest(GQLBase):
    file: ArtifactFragmentCurrentManifestFile


class ArtifactFragmentCurrentManifestFile(GQLBase):
    direct_url: str = Field(alias="directUrl")


class ArtifactFragmentTags(GQLBase):
    name: str


class ArtifactPortfolioTypeFields(GQLBase):
    typename__: Typename[Literal["ArtifactPortfolio"]] = "ArtifactPortfolio"
    id: GQLId
    name: str


class ArtifactSequenceTypeFields(GQLBase):
    typename__: Typename[Literal["ArtifactSequence"]] = "ArtifactSequence"
    id: GQLId
    name: str


class ArtifactTypeFragment(GQLBase):
    typename__: Typename[Literal["ArtifactType"]] = "ArtifactType"
    id: GQLId
    name: str
    description: Optional[str]
    created_at: str = Field(alias="createdAt")


class ArtifactTypesFragment(GQLBase):
    edges: List[ArtifactTypesFragmentEdges]
    page_info: ArtifactTypesFragmentPageInfo = Field(alias="pageInfo")


class ArtifactTypesFragmentEdges(GQLBase):
    node: Optional[ArtifactTypeFragment]
    cursor: str


class ArtifactTypesFragmentPageInfo(GQLBase):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


class ArtifactsFragment(GQLBase):
    total_count: int = Field(alias="totalCount")
    edges: List[ArtifactsFragmentEdges]
    page_info: ArtifactsFragmentPageInfo = Field(alias="pageInfo")


class ArtifactsFragmentEdges(GQLBase):
    node: ArtifactFragment
    version: str
    cursor: str


class ArtifactsFragmentPageInfo(GQLBase):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


class FilesFragment(GQLBase):
    edges: List[FilesFragmentEdges]
    page_info: FilesFragmentPageInfo = Field(alias="pageInfo")


class FilesFragmentEdges(GQLBase):
    node: Optional[FilesFragmentEdgesNode]
    cursor: str


class FilesFragmentEdgesNode(GQLBase):
    id: GQLId
    name: str
    url: Optional[str]
    size_bytes: Any = Field(alias="sizeBytes")
    storage_path: Optional[str] = Field(alias="storagePath", default=None)
    mimetype: Optional[str]
    updated_at: Optional[str] = Field(alias="updatedAt")
    digest: Optional[str]
    md_5: Optional[str] = Field(alias="md5")
    direct_url: str = Field(alias="directUrl")


class FilesFragmentPageInfo(GQLBase):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


ArtifactCollectionsFragment.model_rebuild()
ArtifactCollectionsFragmentEdges.model_rebuild()
ArtifactCollectionsFragmentEdgesNode.model_rebuild()
ArtifactCollectionsFragmentPageInfo.model_rebuild()
ArtifactFragment.model_rebuild()
ArtifactFragmentAliases.model_rebuild()
ArtifactFragmentAliasesArtifactCollection.model_rebuild()
ArtifactFragmentAliasesArtifactCollectionProject.model_rebuild()
ArtifactFragmentArtifactSequence.model_rebuild()
ArtifactFragmentArtifactSequenceProject.model_rebuild()
ArtifactFragmentArtifactType.model_rebuild()
ArtifactFragmentCurrentManifest.model_rebuild()
ArtifactFragmentCurrentManifestFile.model_rebuild()
ArtifactFragmentTags.model_rebuild()
ArtifactPortfolioTypeFields.model_rebuild()
ArtifactSequenceTypeFields.model_rebuild()
ArtifactTypeFragment.model_rebuild()
ArtifactTypesFragment.model_rebuild()
ArtifactTypesFragmentEdges.model_rebuild()
ArtifactTypesFragmentPageInfo.model_rebuild()
ArtifactsFragment.model_rebuild()
ArtifactsFragmentEdges.model_rebuild()
ArtifactsFragmentPageInfo.model_rebuild()
FilesFragment.model_rebuild()
FilesFragmentEdges.model_rebuild()
FilesFragmentEdgesNode.model_rebuild()
FilesFragmentPageInfo.model_rebuild()
ArtifactFragment.model_rebuild()
ArtifactTypeFragment.model_rebuild()



from __future__ import annotations

import atexit
import os
from typing import Callable

from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_server_pb2 as spb
from wandb.proto import wandb_settings_pb2
from wandb.sdk import wandb_settings
from wandb.sdk.interface.interface import InterfaceBase
from wandb.sdk.interface.interface_sock import InterfaceSock
from wandb.sdk.interface.router_sock import MessageSockRouter
from wandb.sdk.lib import service_token
from wandb.sdk.lib.exit_hooks import ExitHooks
from wandb.sdk.lib.sock_client import SockClient, SockClientClosedError
from wandb.sdk.mailbox import HandleAbandonedError, Mailbox, MailboxClosedError
from wandb.sdk.service import service


class WandbServiceNotOwnedError(Exception):
    """Raised when the current process does not own the service process."""


class WandbServiceConnectionError(Exception):
    """Raised on failure to connect to the service process."""


class WandbAttachFailedError(Exception):
    """Raised if attaching to a run fails."""


def connect_to_service(
    settings: wandb_settings.Settings,
) -> ServiceConnection:
    """Connects to the service process, starting one up if necessary."""
    conn = _try_connect_to_existing_service()
    if conn:
        return conn

    return _start_and_connect_service(settings)


def _try_connect_to_existing_service() -> ServiceConnection | None:
    """Attempts to connect to an existing service process."""
    token = service_token.get_service_token()
    if not token:
        return None

    # Only localhost sockets are supported below.
    assert token.host == "localhost"
    client = SockClient()

    try:
        # TODO: This may block indefinitely if the service is unhealthy.
        client.connect(token.port)

    except Exception as e:
        raise WandbServiceConnectionError(
            "Failed to connect to internal service."
        ) from e

    return ServiceConnection(client=client, proc=None)


def _start_and_connect_service(
    settings: wandb_settings.Settings,
) -> ServiceConnection:
    """Starts a service process and returns a connection to it.

    An atexit hook is registered to tear down the service process and wait for
    it to complete. The hook does not run in processes started using the
    multiprocessing module.
    """
    proc = service._Service(settings)
    proc.start()

    port = proc.sock_port
    assert port
    client = SockClient()
    client.connect(port)

    service_token.set_service_token(
        parent_pid=os.getpid(),
        transport="tcp",
        host="localhost",
        port=port,
    )

    hooks = ExitHooks()
    hooks.hook()

    def teardown_atexit():
        conn.teardown(hooks.exit_code)

    conn = ServiceConnection(
        client=client,
        proc=proc,
        cleanup=lambda: atexit.unregister(teardown_atexit),
    )

    atexit.register(teardown_atexit)

    return conn


class ServiceConnection:
    """A connection to the W&B internal service process."""

    def __init__(
        self,
        client: SockClient,
        proc: service._Service | None,
        cleanup: Callable[[], None] | None = None,
    ):
        """Returns a new ServiceConnection.

        Args:
            mailbox: The mailbox to use for all communication over the socket.
            router: A handle to the thread that reads from the socket and
                updates the mailbox.
            client: A socket that's connected to the service.
            proc: The service process if we own it, or None otherwise.
            cleanup: A callback to run on teardown before doing anything.
        """
        self._client = client
        self._proc = proc
        self._torn_down = False
        self._cleanup = cleanup

        self._mailbox = Mailbox()
        self._router = MessageSockRouter(self._client, self._mailbox)

    def make_interface(self, stream_id: str) -> InterfaceBase:
        """Returns an interface for communicating with the service."""
        return InterfaceSock(self._client, self._mailbox, stream_id=stream_id)

    def send_record(self, record: pb.Record) -> None:
        """Sends data to the service."""
        self._client.send_record_publish(record)

    def inform_init(
        self,
        settings: wandb_settings_pb2.Settings,
        run_id: str,
    ) -> None:
        """Sends an init request to the service."""
        request = spb.ServerInformInitRequest()
        request.settings.CopyFrom(settings)
        request._info.stream_id = run_id
        self._client.send_server_request(spb.ServerRequest(inform_init=request))

    def inform_finish(self, run_id: str) -> None:
        """Sends an finish request to the service."""
        request = spb.ServerInformFinishRequest()
        request._info.stream_id = run_id
        self._client.send_server_request(spb.ServerRequest(inform_finish=request))

    def inform_attach(
        self,
        attach_id: str,
    ) -> wandb_settings_pb2.Settings:
        """Sends an attach request to the service.

        Raises a WandbAttachFailedError if attaching is not possible.
        """
        request = spb.ServerRequest()
        request.inform_attach._info.stream_id = attach_id

        try:
            handle = self._mailbox.require_response(request)
            self._client.send_server_request(request)
            response = handle.wait_or(timeout=10)
            return response.inform_attach_response.settings

        except (MailboxClosedError, HandleAbandonedError, SockClientClosedError):
            raise WandbAttachFailedError(
                "Failed to attach: the service process is not running.",
            ) from None

        except TimeoutError:
            raise WandbAttachFailedError(
                "Failed to attach because the run does not belong to"
                " the current service process, or because the service"
                " process is busy (unlikely)."
            ) from None

    def inform_start(
        self,
        settings: wandb_settings_pb2.Settings,
        run_id: str,
    ) -> None:
        """Sends a start request to the service."""
        request = spb.ServerInformStartRequest()
        request.settings.CopyFrom(settings)
        request._info.stream_id = run_id
        self._client.send_server_request(spb.ServerRequest(inform_start=request))

    def teardown(self, exit_code: int) -> int:
        """Shuts down the service process and returns its exit code.

        This may only be called once.

        Returns:
            The exit code of the service process.

        Raises:
            WandbServiceNotOwnedError: If the current process did not start
                the service process.
        """
        if not self._proc:
            raise WandbServiceNotOwnedError(
                "Cannot tear down service started by different process",
            )

        assert not self._torn_down
        self._torn_down = True

        if self._cleanup:
            self._cleanup()

        # Clear the service token to prevent new connections from being made.
        service_token.clear_service_token()

        # Stop reading responses on the socket.
        self._router.join()

        self._client.send_server_request(
            spb.ServerRequest(
                inform_teardown=spb.ServerInformTeardownRequest(
                    exit_code=exit_code,
                )
            ),
        )

        return self._proc.join()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/projects/

__all__ = [
    "DELETE_PROJECT_GQL",
    "FETCH_REGISTRY_GQL",
    "RENAME_PROJECT_GQL",
    "UPSERT_REGISTRY_PROJECT_GQL",
]

FETCH_REGISTRY_GQL = """
query FetchRegistry($name: String, $entityName: String) {
  entity(name: $entityName) {
    project(name: $name) {
      ...RegistryFragment
    }
  }
}

fragment RegistryFragment on Project {
  id
  allowAllArtifactTypesInRegistry
  artifactTypes(includeAll: true) {
    edges {
      node {
        name
      }
    }
  }
  name
  description
  createdAt
  updatedAt
  access
}
"""

RENAME_PROJECT_GQL = """
mutation renameProject($entityName: String!, $oldProjectName: String!, $newProjectName: String!) {
  renameProject(
    input: {entityName: $entityName, oldProjectName: $oldProjectName, newProjectName: $newProjectName}
  ) {
    project {
      name
    }
    inserted
  }
}
"""

UPSERT_REGISTRY_PROJECT_GQL = """
mutation UpsertRegistryProject($description: String, $entityName: String, $name: String, $access: String, $allowAllArtifactTypesInRegistry: Boolean, $artifactTypes: [ArtifactTypeInput!]) {
  upsertModel(
    input: {description: $description, entityName: $entityName, name: $name, access: $access, allowAllArtifactTypesInRegistry: $allowAllArtifactTypesInRegistry, artifactTypes: $artifactTypes}
  ) {
    project {
      ...RegistryFragment
    }
    inserted
  }
}

fragment RegistryFragment on Project {
  id
  allowAllArtifactTypesInRegistry
  artifactTypes(includeAll: true) {
    edges {
      node {
        name
      }
    }
  }
  name
  description
  createdAt
  updatedAt
  access
}
"""

DELETE_PROJECT_GQL = """
mutation deleteProject($id: String!) {
  deleteModel(input: {id: $id}) {
    success
    __typename
  }
}
"""



# Generated by ariadne-codegen
# Source: core/api/graphql/schemas/schema-latest.graphql

from __future__ import annotations

from typing import Any, Optional

from pydantic import Field

from wandb._pydantic import GQLBase


class ArtifactAliasInput(GQLBase):
    artifact_collection_name: str = Field(alias="artifactCollectionName")
    alias: str


class ArtifactCollectionAliasInput(GQLBase):
    alias: str
    entity_name: str = Field(alias="entityName")
    project_name: str = Field(alias="projectName")
    artifact_collection_name: str = Field(alias="artifactCollectionName")


class TagInput(GQLBase):
    tag_category_name: Optional[str] = Field(alias="tagCategoryName", default=None)
    tag_name: str = Field(alias="tagName")
    attributes: Optional[Any] = None



"""Base classes and other customizations for generated pydantic types."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar

from pydantic import BaseModel, ConfigDict, Field, Json, StrictStr
from typing_extensions import Annotated, TypedDict, Unpack, override

from .utils import IS_PYDANTIC_V2, to_json
from .v1_compat import PydanticCompatMixin

if TYPE_CHECKING:
    from pydantic.main import IncEx


class ModelDumpKwargs(TypedDict, total=False):
    """Shared keyword arguments for `BaseModel.model_{dump,dump_json}`."""

    include: IncEx | None
    exclude: IncEx | None
    context: dict[str, Any] | None
    by_alias: bool | None
    exclude_unset: bool
    exclude_defaults: bool
    exclude_none: bool
    round_trip: bool
    warnings: bool | Literal["none", "warn", "error"]
    fallback: Callable[[Any], Any] | None
    serialize_as_any: bool


#: Custom overrides of default kwargs for `BaseModel.model_{dump,dump_json}`.
MODEL_DUMP_DEFAULTS = ModelDumpKwargs(
    by_alias=True,  # Always serialize with aliases (e.g. camelCase names)
    round_trip=True,  # Ensure serialized values remain valid inputs for deserialization
)


# v1-compatible base class for pydantic types.
class CompatBaseModel(PydanticCompatMixin, BaseModel):
    __doc__ = None  # Prevent subclasses from inheriting the BaseModel docstring


# Base class for all GraphQL-generated types.
# Omitted from docstring to avoid inclusion in generated docs.
class GQLBase(CompatBaseModel):
    model_config = ConfigDict(
        populate_by_name=True,  # Discouraged in pydantic v2.11+, will be deprecated in v3
        validate_by_name=True,  # Introduced in pydantic v2.11
        validate_by_alias=True,  # Introduced in pydantic v2.11
        serialize_by_alias=True,  # Introduced in pydantic v2.11
        validate_assignment=True,
        validate_default=True,
        use_attribute_docstrings=True,
        from_attributes=True,
        revalidate_instances="always",
        protected_namespaces=(),  # Some GraphQL fields may begin with "model_"
    )

    @override
    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "json",  # NOTE: changed default
        **kwargs: Unpack[ModelDumpKwargs],
    ) -> dict[str, Any]:
        kwargs = {**MODEL_DUMP_DEFAULTS, **kwargs}
        return super().model_dump(mode=mode, **kwargs)

    @override
    def model_dump_json(
        self,
        *,
        indent: int | None = None,
        **kwargs: Unpack[ModelDumpKwargs],
    ) -> str:
        kwargs = {**MODEL_DUMP_DEFAULTS, **kwargs}
        return super().model_dump_json(indent=indent, **kwargs)


# ------------------------------------------------------------------------------
# Reusable annotations for field types
T = TypeVar("T")

if IS_PYDANTIC_V2 or TYPE_CHECKING:
    GQLId = Annotated[
        StrictStr,
        Field(repr=False, frozen=True),
    ]
else:
    # FIXME: Find a way to fix this for pydantic v1, which doesn't like when
    # `Field(...)` used in the field assignment AND `Annotated[...]`.
    # This is a problem for codegen, which can currently outputs e.g.
    #
    #   class MyModel(GQLBase):
    #       my_id: GQLId = Field(alias="myID")
    #
    GQLId = StrictStr  # type: ignore[misc]

Typename = Annotated[
    T,
    Field(repr=False, frozen=True, alias="__typename"),
]


def ensure_json(v: Any) -> Any:
    """In case the incoming value isn't serialized JSON, reserialize it.

    This lets us use `Json[...]` fields with values that are already deserialized.
    """
    # NOTE: Assumes that the deserialized type is not itself a string.
    # Revisit this if we need to support deserialized types that are str/bytes.
    return v if isinstance(v, (str, bytes)) else to_json(v)


if IS_PYDANTIC_V2 or TYPE_CHECKING:
    from pydantic import BeforeValidator, PlainSerializer

    SerializedToJson = Annotated[
        Json[T],
        # Allow lenient instantiation/validation: incoming data may already be deserialized.
        BeforeValidator(ensure_json),
        PlainSerializer(to_json),
    ]
else:
    # FIXME: Restore, modify, or replace this later after ensuring pydantic v1 compatibility.
    SerializedToJson = Json[T]  # type: ignore[misc]



import codecs
import json
import os

from wandb import util
from wandb.sdk.lib import runid

from . import _dtypes
from ._private import MEDIA_TMP
from .base_types.media import Media


class Bokeh(Media):
    """Wandb class for Bokeh plots.

    Args:
        val: Bokeh plot
    """

    _log_type = "bokeh-file"

    def __init__(self, data_or_path):
        super().__init__()
        bokeh = util.get_module("bokeh", required=True)
        if isinstance(data_or_path, str) and os.path.exists(data_or_path):
            with open(data_or_path) as file:
                b_json = json.load(file)
            self.b_obj = bokeh.document.Document.from_json(b_json)
            self._set_file(data_or_path, is_tmp=False, extension=".bokeh.json")
        elif isinstance(data_or_path, bokeh.model.Model):
            _data = bokeh.document.Document()
            _data.add_root(data_or_path)
            # serialize/deserialize pairing followed by sorting attributes ensures
            # that the file's sha's are equivalent in subsequent calls
            self.b_obj = bokeh.document.Document.from_json(_data.to_json())
            b_json = self.b_obj.to_json()
            if "references" in b_json["roots"]:
                b_json["roots"]["references"].sort(key=lambda x: x["id"])

            tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ".bokeh.json")
            with codecs.open(tmp_path, "w", encoding="utf-8") as fp:
                util.json_dump_safer(b_json, fp)
            self._set_file(tmp_path, is_tmp=True, extension=".bokeh.json")
        elif not isinstance(data_or_path, bokeh.document.Document):
            raise TypeError(
                "Bokeh constructor accepts Bokeh document/model or path to Bokeh json file"
            )

    def get_media_subdir(self):
        return os.path.join("media", "bokeh")

    def to_json(self, run):
        # TODO: (tss) this is getting redundant for all the media objects. We can probably
        # pull this into Media#to_json and remove this type override for all the media types.
        # There are only a few cases where the type is different between artifacts and runs.
        json_dict = super().to_json(run)
        json_dict["_type"] = self._log_type
        return json_dict

    @classmethod
    def from_json(cls, json_obj, source_artifact):
        return cls(source_artifact.get_entry(json_obj["path"]).download())


class _BokehFileType(_dtypes.Type):
    name = "bokeh-file"
    types = [Bokeh]


_dtypes.TypeRegistry.add(_BokehFileType)



# Generated by wandb/proto/wandb_internal_codegen.py.  DO NOT EDIT!

from typing import Literal
DEPRECATED_FEATURES = Literal[
    "keras_callback__data_type",
    "run__mode",
    "run__save_no_args",
    "run__join",
    "plots",
    "run__log_sync",
    "init__config_include_keys",
    "init__config_exclude_keys",
    "keras_callback__save_model",
    "langchain_tracer",
    "artifact__get_path",
    "artifactmanifestentry__name",
    "api__artifact_versions",
    "artifact_collection__change_type",
    "run__define_metric_copy",
    "run_disabled",
    "keras_callback",
    "run__define_metric_best_goal",
    "run__finish_quiet",
    "run__reinit_bool",
    "run__get_url",
    "run__project_name",
    "run__get_project_url",
    "run__get_sweep_url",
]

class Deprecated:
    keras_callback__data_type: DEPRECATED_FEATURES = "keras_callback__data_type"
    run__mode: DEPRECATED_FEATURES = "run__mode"
    run__save_no_args: DEPRECATED_FEATURES = "run__save_no_args"
    run__join: DEPRECATED_FEATURES = "run__join"
    plots: DEPRECATED_FEATURES = "plots"
    run__log_sync: DEPRECATED_FEATURES = "run__log_sync"
    init__config_include_keys: DEPRECATED_FEATURES = "init__config_include_keys"
    init__config_exclude_keys: DEPRECATED_FEATURES = "init__config_exclude_keys"
    keras_callback__save_model: DEPRECATED_FEATURES = "keras_callback__save_model"
    langchain_tracer: DEPRECATED_FEATURES = "langchain_tracer"
    artifact__get_path: DEPRECATED_FEATURES = "artifact__get_path"
    artifactmanifestentry__name: DEPRECATED_FEATURES = "artifactmanifestentry__name"
    api__artifact_versions: DEPRECATED_FEATURES = "api__artifact_versions"
    artifact_collection__change_type: DEPRECATED_FEATURES = "artifact_collection__change_type"
    run__define_metric_copy: DEPRECATED_FEATURES = "run__define_metric_copy"
    run_disabled: DEPRECATED_FEATURES = "run_disabled"
    keras_callback: DEPRECATED_FEATURES = "keras_callback"
    run__define_metric_best_goal: DEPRECATED_FEATURES = "run__define_metric_best_goal"
    run__finish_quiet: DEPRECATED_FEATURES = "run__finish_quiet"
    run__reinit_bool: DEPRECATED_FEATURES = "run__reinit_bool"
    run__get_url: DEPRECATED_FEATURES = "run__get_url"
    run__project_name: DEPRECATED_FEATURES = "run__project_name"
    run__get_project_url: DEPRECATED_FEATURES = "run__get_project_url"
    run__get_sweep_url: DEPRECATED_FEATURES = "run__get_sweep_url"



# Generated by ariadne-codegen
# Source: core/api/graphql/schemas/schema-latest.graphql

from __future__ import annotations

from typing import Optional

from wandb._pydantic import GQLBase


class ArtifactTypeInput(GQLBase):
    name: str
    description: Optional[str] = None



"""Public API: registries search."""

import json
from typing import TYPE_CHECKING, Any, Dict, Optional

if TYPE_CHECKING:
    from wandb_gql import Client

from wandb_gql import gql

import wandb
from wandb.apis.paginator import Paginator
from wandb.apis.public.artifacts import ArtifactCollection
from wandb.apis.public.registries.utils import _ensure_registry_prefix_on_names
from wandb.sdk.artifacts._graphql_fragments import (
    _gql_artifact_fragment,
    _gql_registry_fragment,
)


class Registries(Paginator):
    """Iterator that returns Registries."""

    QUERY = gql(
        """
        query Registries($organization: String!, $filters: JSONString, $cursor: String, $perPage: Int) {
            organization(name: $organization) {
                orgEntity {
                    name
                    projects(filters: $filters, after: $cursor, first: $perPage) {
                        pageInfo {
                            endCursor
                            hasNextPage
                        }
                        edges {
                            node {
                                ...RegistryFragment
                            }
                        }
                    }
                }
            }
        }
        """
        + _gql_registry_fragment()
    )

    def __init__(
        self,
        client: "Client",
        organization: str,
        filter: Optional[Dict[str, Any]] = None,
        per_page: Optional[int] = 100,
    ):
        self.client = client
        self.organization = organization
        self.filter = _ensure_registry_prefix_on_names(filter or {})
        variables = {
            "organization": organization,
            "filters": json.dumps(self.filter),
        }

        super().__init__(client, variables, per_page)

    def __bool__(self):
        return bool(self.objects)

    def __next__(self):
        # Implement custom next since its possible to load empty pages because of auth
        self.index += 1
        while len(self.objects) <= self.index:
            if not self._load_page():
                raise StopIteration
        return self.objects[self.index]

    def collections(self, filter: Optional[Dict[str, Any]] = None) -> "Collections":
        return Collections(
            self.client,
            self.organization,
            registry_filter=self.filter,
            collection_filter=filter,
        )

    def versions(self, filter: Optional[Dict[str, Any]] = None) -> "Versions":
        return Versions(
            self.client,
            self.organization,
            registry_filter=self.filter,
            collection_filter=None,
            artifact_filter=filter,
        )

    @property
    def length(self):
        if self.last_response:
            return len(
                self.last_response["organization"]["orgEntity"]["projects"]["edges"]
            )
        else:
            return None

    @property
    def more(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"]["projects"][
                "pageInfo"
            ]["hasNextPage"]
        else:
            return True

    @property
    def cursor(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"]["projects"][
                "pageInfo"
            ]["endCursor"]
        else:
            return None

    def convert_objects(self):
        if not self.last_response:
            return []
        if (
            not self.last_response["organization"]
            or not self.last_response["organization"]["orgEntity"]
        ):
            raise ValueError(
                f"Organization '{self.organization}' not found. Please verify the organization name is correct"
            )

        from wandb.apis.public.registries.registry import Registry

        return [
            Registry(
                self.client,
                self.organization,
                self.last_response["organization"]["orgEntity"]["name"],
                r["node"]["name"],
                r["node"],
            )
            for r in self.last_response["organization"]["orgEntity"]["projects"][
                "edges"
            ]
        ]


class Collections(Paginator):
    """Iterator that returns Artifact collections in the Registry."""

    QUERY = gql(
        """
        query Collections(
            $organization: String!,
            $registryFilter: JSONString,
            $collectionFilter: JSONString,
            $collectionTypes: [ArtifactCollectionType!],
            $cursor: String,
            $perPage: Int
        ) {
            organization(name: $organization) {
                orgEntity {
                    name
                    artifactCollections(
                        projectFilters: $registryFilter,
                        filters: $collectionFilter,
                        collectionTypes: $collectionTypes,
                        after: $cursor,
                        first: $perPage
                    ) {
                        totalCount
                        pageInfo {
                            endCursor
                            hasNextPage
                        }
                        edges {
                            cursor
                            node {
                                id
                                name
                                description
                                createdAt
                                tags {
                                    edges {
                                        node {
                                            name
                                        }
                                    }
                                }
                                project {
                                    name
                                    entity {
                                        name
                                    }
                                }
                                defaultArtifactType {
                                    name
                                }
                                aliases {
                                    edges {
                                        node {
                                            alias
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        """
    )

    def __init__(
        self,
        client: "Client",
        organization: str,
        registry_filter: Optional[Dict[str, Any]] = None,
        collection_filter: Optional[Dict[str, Any]] = None,
        per_page: Optional[int] = 100,
    ):
        self.client = client
        self.organization = organization
        self.registry_filter = registry_filter
        self.collection_filter = collection_filter or {}

        variables = {
            "registryFilter": (
                json.dumps(self.registry_filter) if self.registry_filter else None
            ),
            "collectionFilter": (
                json.dumps(self.collection_filter) if self.collection_filter else None
            ),
            "organization": self.organization,
            "collectionTypes": ["PORTFOLIO"],
            "perPage": per_page,
        }

        super().__init__(client, variables, per_page)

    def __bool__(self):
        return len(self) > 0 or len(self.objects) > 0

    def __next__(self):
        # Implement custom next since its possible to load empty pages because of auth
        self.index += 1
        while len(self.objects) <= self.index:
            if not self._load_page():
                raise StopIteration
        return self.objects[self.index]

    def versions(self, filter: Optional[Dict[str, Any]] = None) -> "Versions":
        return Versions(
            self.client,
            self.organization,
            registry_filter=self.registry_filter,
            collection_filter=self.collection_filter,
            artifact_filter=filter,
        )

    @property
    def length(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["totalCount"]
        else:
            return None

    @property
    def more(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["pageInfo"]["hasNextPage"]
        else:
            return True

    @property
    def cursor(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["pageInfo"]["endCursor"]
        else:
            return None

    def convert_objects(self):
        if not self.last_response:
            return []
        if (
            not self.last_response["organization"]
            or not self.last_response["organization"]["orgEntity"]
        ):
            raise ValueError(
                f"Organization '{self.organization}' not found. Please verify the organization name is correct"
            )

        return [
            ArtifactCollection(
                self.client,
                r["node"]["project"]["entity"]["name"],
                r["node"]["project"]["name"],
                r["node"]["name"],
                r["node"]["defaultArtifactType"]["name"],
                self.organization,
                r["node"],
                is_sequence=False,
            )
            for r in self.last_response["organization"]["orgEntity"][
                "artifactCollections"
            ]["edges"]
        ]


class Versions(Paginator):
    """Iterator that returns Artifact versions in the Registry."""

    def __init__(
        self,
        client: "Client",
        organization: str,
        registry_filter: Optional[Dict[str, Any]] = None,
        collection_filter: Optional[Dict[str, Any]] = None,
        artifact_filter: Optional[Dict[str, Any]] = None,
        per_page: int = 100,
    ):
        self.client = client
        self.organization = organization
        self.registry_filter = registry_filter
        self.collection_filter = collection_filter
        self.artifact_filter = artifact_filter or {}
        self.QUERY = gql(
            """
            query Versions(
                $organization: String!,
                $registryFilter: JSONString,
                $collectionFilter: JSONString,
                $artifactFilter: JSONString,
                $cursor: String,
                $perPage: Int
            ) {
                organization(name: $organization) {
                    orgEntity {
                        name
                        artifactMemberships(
                            projectFilters: $registryFilter,
                            collectionFilters: $collectionFilter,
                            filters: $artifactFilter,
                            after: $cursor,
                            first: $perPage
                        ) {
                            pageInfo {
                                endCursor
                                hasNextPage
                            }
                            edges {
                                node {
                                    artifactCollection {
                                        project {
                                            name
                                            entity {
                                                name
                                            }
                                        }
                                        name
                                    }
                                    versionIndex
                                    artifact {
                                        ...ArtifactFragment
                                    }
                                    aliases {
                                        alias
                                    }
                                }
                            }
                        }
                    }
                }
            }
            """
            + _gql_artifact_fragment(include_aliases=False)
        )

        variables = {
            "registryFilter": (
                json.dumps(self.registry_filter) if self.registry_filter else None
            ),
            "collectionFilter": (
                json.dumps(self.collection_filter) if self.collection_filter else None
            ),
            "artifactFilter": (
                json.dumps(self.artifact_filter) if self.artifact_filter else None
            ),
            "organization": self.organization,
        }

        super().__init__(client, variables, per_page)

    def __next__(self):
        # Implement custom next since its possible to load empty pages because of auth
        self.index += 1
        while len(self.objects) <= self.index:
            if not self._load_page():
                raise StopIteration
        return self.objects[self.index]

    def __bool__(self):
        return len(self) > 0 or len(self.objects) > 0

    @property
    def length(self):
        if self.last_response:
            return len(
                self.last_response["organization"]["orgEntity"]["artifactMemberships"][
                    "edges"
                ]
            )
        else:
            return None

    @property
    def more(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactMemberships"
            ]["pageInfo"]["hasNextPage"]
        else:
            return True

    @property
    def cursor(self):
        if self.last_response:
            return self.last_response["organization"]["orgEntity"][
                "artifactMemberships"
            ]["pageInfo"]["endCursor"]
        else:
            return None

    def convert_objects(self):
        if not self.last_response:
            return []
        if (
            not self.last_response["organization"]
            or not self.last_response["organization"]["orgEntity"]
        ):
            raise ValueError(
                f"Organization '{self.organization}' not found. Please verify the organization name is correct"
            )

        artifacts = (
            wandb.Artifact._from_attrs(
                a["node"]["artifactCollection"]["project"]["entity"]["name"],
                a["node"]["artifactCollection"]["project"]["name"],
                a["node"]["artifactCollection"]["name"]
                + ":v"
                + str(a["node"]["versionIndex"]),
                a["node"]["artifact"],
                self.client,
                [alias["alias"] for alias in a["node"]["aliases"]],
            )
            for a in self.last_response["organization"]["orgEntity"][
                "artifactMemberships"
            ]["edges"]
        )
        return artifacts



from __future__ import annotations

from typing import Iterable, TypeVar

T = TypeVar("T")


def one(
    iterable: Iterable[T],
    too_short: Exception | None = None,
    too_long: Exception | None = None,
) -> T:
    """Return the only item in the iterable.

    Note:
        This is intended **only** as an internal helper/convenience function,
        and its implementation is directly adapted from `more_itertools.one`.
        Users needing similar functionality are strongly encouraged to use
        that library instead:
        https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.one

    Args:
        iterable: The iterable to get the only item from.
        too_short: Custom exception to raise if the iterable has no items.
        too_long: Custom exception to raise if the iterable has multiple items.

    Raises:
        ValueError or `too_short`: If the iterable has no items.
        ValueError or `too_long`: If the iterable has multiple items.
    """
    # For a general iterable, avoid inadvertently iterating through all values,
    # which may be costly or impossible (e.g. if infinite).  Only check that:

    # ... the first item exists
    it = iter(iterable)
    try:
        obj = next(it)
    except StopIteration as e:
        raise too_short or ValueError("Expected 1 item in iterable, got 0") from e

    # ...the second item doesn't
    try:
        _ = next(it)
    except StopIteration:
        return obj
    raise too_long or ValueError("Expected 1 item in iterable, got multiple")



import hashlib
import os
from typing import Optional

from wandb import util
from wandb.sdk.lib import filesystem, runid

from . import _dtypes
from ._private import MEDIA_TMP
from .base_types.media import BatchableMedia


class Audio(BatchableMedia):
    """Wandb class for audio clips.

    Args:
        data_or_path: (string or numpy array) A path to an audio file
            or a numpy array of audio data.
        sample_rate: (int) Sample rate, required when passing in raw
            numpy array of audio data.
        caption: (string) Caption to display with audio.
    """

    _log_type = "audio-file"

    def __init__(self, data_or_path, sample_rate=None, caption=None):
        """Accept a path to an audio file or a numpy array of audio data."""
        super().__init__(caption=caption)
        self._duration = None
        self._sample_rate = sample_rate

        if isinstance(data_or_path, str):
            if self.path_is_reference(data_or_path):
                self._path = data_or_path
                self._sha256 = hashlib.sha256(data_or_path.encode("utf-8")).hexdigest()
                self._is_tmp = False
            else:
                self._set_file(data_or_path, is_tmp=False)
        else:
            if sample_rate is None:
                raise ValueError(
                    'Argument "sample_rate" is required when instantiating wandb.Audio with raw data.'
                )

            soundfile = util.get_module(
                "soundfile",
                required='Raw audio requires the soundfile package. To get it, run "pip install soundfile"',
            )

            tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ".wav")
            soundfile.write(tmp_path, data_or_path, sample_rate)
            self._duration = len(data_or_path) / float(sample_rate)

            self._set_file(tmp_path, is_tmp=True)

    @classmethod
    def get_media_subdir(cls):
        return os.path.join("media", "audio")

    @classmethod
    def from_json(cls, json_obj, source_artifact):
        return cls(
            source_artifact.get_entry(json_obj["path"]).download(),
            caption=json_obj["caption"],
        )

    def bind_to_run(
        self, run, key, step, id_=None, ignore_copy_err: Optional[bool] = None
    ):
        if self.path_is_reference(self._path):
            raise ValueError(
                "Audio media created by a reference to external storage cannot currently be added to a run"
            )

        return super().bind_to_run(run, key, step, id_, ignore_copy_err)

    def to_json(self, run):
        json_dict = super().to_json(run)
        json_dict.update(
            {
                "_type": self._log_type,
            }
        )
        return json_dict

    @classmethod
    def seq_to_json(cls, seq, run, key, step):
        audio_list = list(seq)

        util.get_module(
            "soundfile",
            required="wandb.Audio requires the soundfile package. To get it, run: pip install soundfile",
        )
        base_path = os.path.join(run.dir, "media", "audio")
        filesystem.mkdir_exists_ok(base_path)
        meta = {
            "_type": "audio",
            "count": len(audio_list),
            "audio": [a.to_json(run) for a in audio_list],
        }
        sample_rates = cls.sample_rates(audio_list)
        if sample_rates:
            meta["sampleRates"] = sample_rates
        durations = cls.durations(audio_list)
        if durations:
            meta["durations"] = durations
        captions = cls.captions(audio_list)
        if captions:
            meta["captions"] = captions

        return meta

    @classmethod
    def durations(cls, audio_list):
        return [a._duration for a in audio_list]

    @classmethod
    def sample_rates(cls, audio_list):
        return [a._sample_rate for a in audio_list]

    @classmethod
    def captions(cls, audio_list):
        captions = [a._caption for a in audio_list]
        if all(c is None for c in captions):
            return False
        else:
            return ["" if c is None else c for c in captions]

    def resolve_ref(self):
        if self.path_is_reference(self._path):
            # this object was already created using a ref:
            return self._path
        source_artifact = self._artifact_source.artifact

        resolved_name = source_artifact._local_path_to_name(self._path)
        if resolved_name is not None:
            target_entry = source_artifact.manifest.get_entry_by_path(resolved_name)
            if target_entry is not None:
                return target_entry.ref

        return None

    def __eq__(self, other):
        if self.path_is_reference(self._path) or self.path_is_reference(other._path):
            # one or more of these objects is an unresolved reference -- we'll compare
            # their reference paths instead of their SHAs:
            return (
                self.resolve_ref() == other.resolve_ref()
                and self._caption == other._caption
            )

        return super().__eq__(other) and self._caption == other._caption

    def __ne__(self, other):
        return not self.__eq__(other)


class _AudioFileType(_dtypes.Type):
    name = "audio-file"
    types = [Audio]


_dtypes.TypeRegistry.add(_AudioFileType)



"""Artifact manifest."""

from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

from wandb.sdk.internal.internal_api import Api as InternalApi
from wandb.sdk.lib.hashutil import HexMD5

if TYPE_CHECKING:
    from wandb.sdk.artifacts.artifact_manifest_entry import ArtifactManifestEntry
    from wandb.sdk.artifacts.storage_policy import StoragePolicy


class ArtifactManifest:
    entries: dict[str, ArtifactManifestEntry]

    @classmethod
    def from_manifest_json(
        cls, manifest_json: dict, api: InternalApi | None = None
    ) -> ArtifactManifest:
        if "version" not in manifest_json:
            raise ValueError("Invalid manifest format. Must contain version field.")
        version = manifest_json["version"]
        for sub in cls.__subclasses__():
            if sub.version() == version:
                return sub.from_manifest_json(manifest_json, api=api)
        raise ValueError("Invalid manifest version.")

    @classmethod
    def version(cls) -> int:
        raise NotImplementedError

    def __init__(
        self,
        storage_policy: StoragePolicy,
        entries: Mapping[str, ArtifactManifestEntry] | None = None,
    ) -> None:
        self.storage_policy = storage_policy
        self.entries = dict(entries) if entries else {}

    def __len__(self) -> int:
        return len(self.entries)

    def to_manifest_json(self) -> dict:
        raise NotImplementedError

    def digest(self) -> HexMD5:
        raise NotImplementedError

    def add_entry(self, entry: ArtifactManifestEntry, overwrite: bool = False) -> None:
        path = entry.path
        if not overwrite:
            prev_entry = self.entries.get(path)
            if prev_entry and (entry.digest != prev_entry.digest):
                raise ValueError(f"Cannot add the same path twice: {path!r}")
        self.entries[path] = entry

    def remove_entry(self, entry: ArtifactManifestEntry) -> None:
        try:
            del self.entries[entry.path]
        except LookupError:
            raise FileNotFoundError(f"Cannot remove missing entry: '{entry.path}'")

    def get_entry_by_path(self, path: str) -> ArtifactManifestEntry | None:
        return self.entries.get(path)

    def get_entries_in_directory(self, directory: str) -> list[ArtifactManifestEntry]:
        return [
            self.entries[entry_key]
            for entry_key in self.entries
            if entry_key.startswith(
                directory + "/"
            )  # entries use forward slash even for windows
        ]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Literal, Optional

from pydantic import Field

from wandb._pydantic import GQLBase, Typename

from .enums import ArtifactCollectionState


class DeleteArtifactPortfolio(GQLBase):
    delete_artifact_portfolio: Optional[
        DeleteArtifactPortfolioDeleteArtifactPortfolio
    ] = Field(alias="deleteArtifactPortfolio")


class DeleteArtifactPortfolioDeleteArtifactPortfolio(GQLBase):
    artifact_collection: DeleteArtifactPortfolioDeleteArtifactPortfolioArtifactCollection = Field(
        alias="artifactCollection"
    )


class DeleteArtifactPortfolioDeleteArtifactPortfolioArtifactCollection(GQLBase):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    state: ArtifactCollectionState


DeleteArtifactPortfolio.model_rebuild()
DeleteArtifactPortfolioDeleteArtifactPortfolio.model_rebuild()



"""InterfaceQueue - Derived from InterfaceShared using queues to send to internal thread.

See interface.py for how interface classes relate to each other.

"""

import logging
from multiprocessing.process import BaseProcess
from typing import TYPE_CHECKING, Optional

from wandb.sdk.mailbox import Mailbox

from .interface_shared import InterfaceShared

if TYPE_CHECKING:
    from queue import Queue

    from wandb.proto import wandb_internal_pb2 as pb


logger = logging.getLogger("wandb")


class InterfaceQueue(InterfaceShared):
    def __init__(
        self,
        record_q: Optional["Queue[pb.Record]"] = None,
        result_q: Optional["Queue[pb.Result]"] = None,
        process: Optional[BaseProcess] = None,
        mailbox: Optional[Mailbox] = None,
    ) -> None:
        self.record_q = record_q
        self.result_q = result_q
        self._process = process
        super().__init__(mailbox=mailbox)

    def _publish(self, record: "pb.Record", local: Optional[bool] = None) -> None:
        if self._process and not self._process.is_alive():
            raise Exception("The wandb backend process has shutdown")
        if local:
            record.control.local = local
        if self.record_q:
            self.record_q.put(record)



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import ArtifactFragment


class RunInputArtifacts(GQLBase):
    project: Optional[RunInputArtifactsProject]


class RunInputArtifactsProject(GQLBase):
    run: Optional[RunInputArtifactsProjectRun]


class RunInputArtifactsProjectRun(GQLBase):
    input_artifacts: Optional[RunInputArtifactsProjectRunInputArtifacts] = Field(
        alias="inputArtifacts"
    )


class RunInputArtifactsProjectRunInputArtifacts(GQLBase):
    total_count: int = Field(alias="totalCount")
    edges: List[RunInputArtifactsProjectRunInputArtifactsEdges]
    page_info: RunInputArtifactsProjectRunInputArtifactsPageInfo = Field(
        alias="pageInfo"
    )


class RunInputArtifactsProjectRunInputArtifactsEdges(GQLBase):
    node: Optional[ArtifactFragment]
    cursor: str


class RunInputArtifactsProjectRunInputArtifactsPageInfo(GQLBase):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


RunInputArtifacts.model_rebuild()
RunInputArtifactsProject.model_rebuild()
RunInputArtifactsProjectRun.model_rebuild()
RunInputArtifactsProjectRunInputArtifacts.model_rebuild()
RunInputArtifactsProjectRunInputArtifactsEdges.model_rebuild()



from __future__ import annotations

import base64
import hashlib
import mmap
import sys
from typing import TYPE_CHECKING, NewType

from wandb.sdk.lib.paths import StrPath

if TYPE_CHECKING:
    import _hashlib  # type: ignore[import-not-found]

ETag = NewType("ETag", str)
HexMD5 = NewType("HexMD5", str)
B64MD5 = NewType("B64MD5", str)


def _md5(data: bytes = b"") -> _hashlib.HASH:
    """Allow FIPS-compliant md5 hash when supported."""
    if sys.version_info >= (3, 9):
        return hashlib.md5(data, usedforsecurity=False)
    else:
        return hashlib.md5(data)


def md5_string(string: str) -> B64MD5:
    return _b64_from_hasher(_md5(string.encode("utf-8")))


def _b64_from_hasher(hasher: _hashlib.HASH) -> B64MD5:
    return B64MD5(base64.b64encode(hasher.digest()).decode("ascii"))


def b64_to_hex_id(string: B64MD5) -> HexMD5:
    return HexMD5(base64.standard_b64decode(string).hex())


def hex_to_b64_id(encoded_string: str | bytes) -> B64MD5:
    if isinstance(encoded_string, bytes):
        encoded_string = encoded_string.decode("utf-8")
    as_str = bytes.fromhex(encoded_string)
    return B64MD5(base64.standard_b64encode(as_str).decode("utf-8"))


def md5_file_b64(*paths: StrPath) -> B64MD5:
    return _b64_from_hasher(_md5_file_hasher(*paths))


def md5_file_hex(*paths: StrPath) -> HexMD5:
    return HexMD5(_md5_file_hasher(*paths).hexdigest())


_KB: int = 1_024
_CHUNKSIZE: int = 128 * _KB
"""Chunk size (in bytes) for iteratively reading from file, if needed."""


def _md5_file_hasher(*paths: StrPath) -> _hashlib.HASH:
    md5_hash = _md5()

    # Note: We use str paths (instead of pathlib.Path objs) for minor perf improvements.
    for path in sorted(map(str, paths)):
        with open(path, "rb") as f:
            try:
                with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ) as mview:
                    md5_hash.update(mview)
            except OSError:
                # This occurs if the mmap-ed file is on a different/mounted filesystem,
                # so we'll fall back on a less performant implementation.

                # Note: At the time of implementation, the walrus operator `:=`
                # is avoided to maintain support for users on python 3.7.
                # Consider revisiting once 3.7 support is no longer needed.
                chunk = f.read(_CHUNKSIZE)
                while chunk:
                    md5_hash.update(chunk)
                    chunk = f.read(_CHUNKSIZE)
            except ValueError:
                # This occurs when mmap-ing an empty file, which can be skipped.
                # See: https://github.com/python/cpython/blob/986a4e1b6fcae7fe7a1d0a26aea446107dd58dd2/Modules/mmapmodule.c#L1589
                pass

    return md5_hash
